"""
Script to check order axe and return axe for scipy
"""
from borea.utils.check.check_header import check_head


def check_order_axe(order_axe: str) -> str:
    """
    Check if code is good and convert for scipy rotation.

    Args:
        order_axe (str): Order of rotation matrix axes.

    Retuns:
        str: Order of ratation matrix axe for scipy.
    """
    list_val = ['o', 'p', 'k']

    bad_head, ms_error_letter, _, symbol = check_head(list(order_axe), list_val)

    misss = set(list_val) - symbol
    if misss != set():
        bad_head = True
        ms_error_letter += f"The letters {misss} are missing.\n"

    ms_error = "Your order axe is not correct.\n"
    ms_error += ms_error_letter
    if bad_head:
        raise ValueError(ms_error)

    return convert_opk_to_xyz(order_axe)


def convert_opk_to_xyz(order_axe: str) -> str:
    """
    Convert for scipy formalism.

    Args:
        order_axe (str): Order of rotation matrix axes.

    Retuns:
        str: Order of ratation matrix axe for scipy.
    """
    letter_conv = {"o": "x", "p": "y", "k": "z"}

    form_xyz = ""
    for i in order_axe:
        form_xyz += letter_conv[i]

    return form_xyz



"""
Photogrammetry worksite to writing dataframe to txt.
"""
import os
from pathlib import Path, PureWindowsPath
import pandas as pd


def write_df_to_txt(name: str, pathreturn: str, df: pd.DataFrame) -> None:
    """
    Writing DataFrame to txt in column.

    Args:
        name (str): Name of the file.
        pathreturn (str): Path to save the file.
        df (pd.DataFrame): DataFrame to save.
    """
    path_txt = os.path.join(Path(PureWindowsPath(pathreturn)), f"{name}.txt")

    name_column = list(df.columns)

    with open(path_txt, "w", encoding="utf-8") as file:
        for _, row in df.iterrows():
            wline = ""
            for id_col in name_column:
                if not isinstance(row[id_col], float):
                    wline += f"{row[id_col]} "
                else:
                    wline += f"{round(row[id_col], 3)} "
            wline += "\n"
            file.write(wline)
        file.close()



"""
A module for manipulating an euclidean projection.
"""
from dataclasses import dataclass
import numpy as np


@dataclass
class EuclideanProj:
    """
    This class represents a Euclidean projection system.

    Args:
        x_central (float): x coordinate of the central point of the Euclidean system.
        y_central (float): y coordinate of the central point of the Euclidean system.

    .. note::
        A Euclidean system is a local Euclidean reference system in which
        the collinear equation is valid.
    """
    x_central: float
    y_central: float

    def __post_init__(self) -> None:
        self.pt_central = np.array([self.x_central, self.y_central, 0])



"""
Main to calculate world coodinate with opk.
"""
# pylint: disable=import-error, wrong-import-position
import argparse
import sys
import os

sys.path.append(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))
from borea.process.p_format.p_read_opk import args_reading_opk, process_args_read_opk  # noqa: E402
from borea.process.p_func.p_image_world import args_image_world, process_image_world  # noqa: E402


def ptfile_image_to_world():
    """
    Converts a image point file into an terrain point file.
    """
    parser = argparse.ArgumentParser(description='Calculate ground coodinate of the image point.')
    # Args for implement ptfile image to world
    parser = args_reading_opk(parser)
    parser = args_image_world(parser)

    args = parser.parse_args()

    # Process to read data
    work = process_args_read_opk(args)
    # Process to transform image to world
    process_image_world(args, work)


if __name__ == "__main__":
    ptfile_image_to_world()



"""
Args of parser to read gcp 3D file.
"""
import argparse
from borea.worksite.worksite import Worksite
from borea.reader.reader_point import read_file_pt


def args_gcp3d(parser: argparse) -> argparse:
    """
    Args to read gcp 3D file.

    Args:
        parser (argparse): Parser to add argument.

    Returns:
        argsparse: Parser with argument.
    """
    parser.add_argument('-g', '--gcp3d',
                        type=str, default=None,
                        help='File path of ground control points in terrain.')
    parser.add_argument('-l', '--head_gcp3d',
                        type=str, default="PTXYZ",
                        help='Header of the file gcp3d.'
                        'e.g. PTXYZ'
                        'S: to ignore the column'
                        'P: name of point'
                        'T: Type of gcp to control.'
                        'X: coordinate x of the shot position'
                        'Y: coordinate y of the shot position'
                        'Z: coordinate z altitude of the shot position'
                        'H: coordinate z height of the shot position')
    return parser


def process_gcp3d(args, work: Worksite) -> Worksite:
    """
    Processing args with data.

    Args:
        args (argparse): Arg to apply on worksite (data).
        work (Worksite): Data.

    Returns:
        Worksite: data
    """
    # Reading GCP 3D
    if args.gcp3d is not None:
        read_file_pt(args.gcp3d, list(args.head_gcp3d.upper()), "gcp3d", work)
        print("GCP reading done.")
        print(f"Number of gcp: {len(work.gcp3d)}")

    return work



"""
A script combining various data conversion and format verification functions.
"""
from typing import Union
import numpy as np


def check_array_transfo(x: Union[np.ndarray, float],
                        y: Union[np.ndarray, float],
                        z: Union[np.ndarray, float] = None) -> tuple:
    """
    Checks the dimension of vectors if dim = 1 outputs the element in a variable.
    x,y,z are in the same dimension and they must be vectors if array
    possible dimension if array : (1,), (n,), (n,1) or (1,n).

    Args:
        x (Union[np.array, float]): Vector x.
        y (Union[np.array, float]): Vector y.
        z (Union[np.array, float]): Vector z.

    Returns:
        tuple: x, y, z.
    """
    if isinstance(x, np.ndarray):  # look if is an array
        t = np.shape(x)
        if len(t) == 1:  # look dim == (n,)
            if t[0] == 1:  # look n == 1
                # takes out the elements
                x = float(x[0])
                y = float(y[0])
                if z is not None:
                    z = float(z[0])

        if len(t) == 2:  # look dim == (n,m)
            if t[0] != 1:  # look n != 1
                # conversion dim (n,1) to (1,n)
                x = x.T
                y = y.T
                if z is not None:
                    z = z.T
            # conversion dim (1,n) to (n,)
            x = np.squeeze(x, axis=0)
            y = np.squeeze(y, axis=0)
            if z is not None:
                z = np.squeeze(z, axis=0)

            # same checkup to the top
            t2 = np.shape(x)
            if len(t2) == 1:
                if t2[0] == 1:
                    x = float(x[0])
                    y = float(y[0])
                    if z is not None:
                        z = float(z[0])

    return x, y, z



"""
A script for verification header list str.
"""


def check_head(header: list, check_letter: list) -> tuple:
    """
    Checks that the header contains all the letters in the check letter.

    Args:
        header (list): List of column type file.
        check_letter (list): List of good letter in header.

    Returns:
        tuple: bad_head, ms_error, head, symbol
    """
    bad_head = False
    ms_error_letter = ""
    symbol = set()
    head = []

    for l_type in header:
        if l_type in check_letter:
            if l_type not in head or l_type == "S":
                head.append(l_type)
                if l_type != "S":
                    symbol.add(l_type)
            else:
                bad_head = True
                ms_error_letter += f"Symbol {l_type} appears several times, "
                ms_error_letter += "this is incorrect, must appear only once\n"
        else:
            bad_head = True
            ms_error_letter += f"Symbol {l_type} is not recognized, "
            ms_error_letter += f"list of symbol recognized {check_letter}\n"

    return bad_head, ms_error_letter, head, symbol


def check_h_z(bad_head: bool, misss: set, ms_error_letter: str) -> tuple:
    """
    Check letter H and Z in header.

    Args:
        bad_head (bool): boolean if the header is false.
        misss (set): Diff between list_letter and symbol.
        ms_error_letter (str): Error message.

    Returns:
        tuple: bad_head, ms_error_letter
    """
    if misss != {"Z"} and misss != {"H"} and misss != set():
        bad_head = True
        if "Z" in misss and "H" in misss:
            miss = misss.copy()
            miss.remove("Z")
            miss.remove("H")
            ms_error_letter += f"The letters 'Z' or 'H' are missing and lettres {miss}.\n"
        elif "Z" in misss:
            misss.remove("Z")
            ms_error_letter += f"The letters {misss} are missing.\n"
        else:
            misss.remove("H")
            ms_error_letter += f"The letters {misss} are missing.\n"

    if misss == set():
        bad_head = True
        ms_error_letter += "The letters Z and H cannot be in the same string.\n"

    return bad_head, ms_error_letter


def get_type_z_and_header(header: list) -> tuple:
    """
    Return type of z, height if H and altitude if Z
    and header with the H replaced by a Z.

    Args:
        header (list): List of column type file.

    Returns:
        tuple: Header and type_z.
    """
    if "H" in header:
        type_z = "height"
        header[header.index('H')] = "Z"
    else:
        type_z = "altitude"

    return header, type_z



"""
Main to transform projection of 3D points.
"""
# pylint: disable=import-error, wrong-import-position, line-too-long
import argparse
import sys
import os

sys.path.append(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))
from borea.process.p_func.p_tf_proj_pt import args_tf_proj_param, process_tf_proj_param  # noqa: E402, E501


def transform_proj_pt():
    """
    Transforms the projection of a point file by pyproj.
    """
    parser = argparse.ArgumentParser(description='Transform coordinate projection of the 3D points'
                                                 ' file.')

    # Args
    parser = args_tf_proj_param(parser)

    args = parser.parse_args()

    # Process data
    process_tf_proj_param(args)


if __name__ == "__main__":
    transform_proj_pt()



"""
Ground Control Point (GCP) class.
"""
from dataclasses import dataclass
import numpy as np


@dataclass
class GCP:
    """
    Ground Control Point class.

    Args:
        name_gcp (str): Name of the gcp.
        code (str): IGN code to differentiate between support points (1, 2, 3)
                    and control points (11, 12, 13)
                    1 means precision in Z, 2 in X and Y and 3 in X, Y, Z.
        coor (numpy.array): Array of ground coordinate [X, Y, Z].
    """
    name_gcp: str
    code: str
    coor: np.ndarray



"""
Module for statistics
"""
import os
import io
from pathlib import Path, PureWindowsPath
import numpy as np
from borea.worksite.worksite import Worksite


class Stat:
    """
    Calculates site statistics.
    """
    def __init__(self, work: Worksite, pathoutput: str, type_point: list = None) -> None:
        """
        Definition of the class Stat.

        Args:
            work (Worksite): The worksite to be treated.
            pathoutput (str): Path to save the file.
            type_point (list): List of type point on which we make the stats.
        """
        self.work = work
        self.pathoutput = Path(PureWindowsPath(pathoutput))

        if type_point is None:
            self.type_point = []
        else:
            self.type_point = type_point

        self.res_world_image = []
        self.res_image_world = []
        self.stat_world_image = {}
        self.stat_image_world = {}

    def main_stat_and_save(self) -> None:
        """
        Calculates and saves all site statistics.
        """
        self.main_stat_world_to_image()
        self.main_stat_image_to_world()
        self.save_stat_txt()

    def main_stat_world_to_image(self) -> None:
        """
        Calculates residual and statistics on control point for world to image function.
        """
        self.stat_world_to_image()
        if self.res_world_image:
            self.stat_world_image = self.stat_list(self.res_world_image)

    def main_stat_image_to_world(self) -> None:
        """
        Calculates residual and statistics on control point for image to world function.
        """
        self.stat_image_to_world()
        if self.res_image_world:
            self.stat_image_world = self.stat_list(self.res_image_world)

    def stat_world_to_image(self) -> None:
        """
        Calculates residual of controle point for world to image.
        residual = image's ground point - calculated image's ground point
        """
        for name_gcp, l_shot in self.work.gcp2d.items():
            try:
                if self.work.gcp3d[name_gcp].code in self.type_point or self.type_point == []:
                    for shot in l_shot:
                        try:
                            img_coor = self.work.shots[shot].gcp2d[name_gcp]
                            img_coor_calculated = self.work.shots[shot].gcp3d[name_gcp]
                            l_data = [[name_gcp, shot], img_coor - img_coor_calculated]
                            self.res_world_image.append(l_data)
                        except KeyError:
                            continue
            except KeyError:
                continue

    def stat_image_to_world(self) -> None:
        """
        Calculates residual of controle point for image to world.
        residual = ground control point - calculated ground control point
        """
        for name_gcp in list(self.work.gcp2d):
            try:
                if self.work.gcp3d[name_gcp].code in self.type_point or self.type_point == []:
                    try:
                        gcp_coor = self.work.gcp3d[name_gcp].coor
                        gcp_coor_calculated = self.work.gcp2d_in_world[name_gcp]
                        l_data = [[name_gcp], gcp_coor - gcp_coor_calculated]
                        self.res_image_world.append(l_data)
                    except KeyError:
                        continue
            except KeyError:
                continue

    def stat_list(self, data_list: list) -> dict:
        """
        Calculates statistics on residual data.
        Min, Max, Median, Mean, Var, Sigma arithmetic and absolute.

        Args:
            data_list (list): List of data.
        """
        dict_output = {}
        data = []
        for ld in data_list:
            data.append(ld[-1])
        data = np.array(data)

        list_stat1 = [[np.amin, np.amax], ["Min", "Max"], [np.argmin, np.argmax]]
        list_stat2 = [[np.median, np.mean, np.var, np.std],
                      ["Median", "Mean", "Var", "Sigma"]]

        for func, name, argfunc in zip(list_stat1[0], list_stat1[1], list_stat1[2]):
            pstat = []
            abspstat = []
            for i, j in zip(argfunc(data, axis=0), argfunc(abs(data), axis=0)):
                if len(data_list[i][0]) == 1:
                    pstat.append(data_list[i][0][0])
                    abspstat.append(data_list[j][0][0])
                else:
                    pstat.append(data_list[i][0])
                    abspstat.append(data_list[j][0])
            dict_output[f"{name}_arith"] = {"val": func(data, axis=0), "data": pstat}
            dict_output[f"{name}_abs"] = {"val": func(abs(data), axis=0), "data": abspstat}

        for func, name in zip(list_stat2[0], list_stat2[1]):
            dict_output[f"{name}_arith"] = np.round(func(data, axis=0), 2)
            dict_output[f"{name}_abs"] = np.round(func(abs(data), axis=0), 2)

        return dict_output

    def save_stat_txt(self) -> None:
        """
        Save calculation statistics in a .txt file.
        """
        path_riw = os.path.join(self.pathoutput, f"Stat_residu_image_to_world_{self.work.name}.txt")
        path_miw = os.path.join(self.pathoutput, f"Stat_metric_image_to_world_{self.work.name}.txt")
        path_rwi = os.path.join(self.pathoutput, f"Stat_residu_world_to_image_{self.work.name}.txt")
        path_mwi = os.path.join(self.pathoutput, f"Stat_metric_world_to_image_{self.work.name}.txt")

        if self.res_image_world:
            try:
                with open(path_riw, "w", encoding="utf-8") as file_riw:
                    file_riw.write("Control point statistics file.\n")
                    file_riw.write("\n")
                    file_riw.write("\n")
                    file_riw.write("Residue of control points from image to terrain function.\n")
                    file_riw.write("residual = ground control point -"
                                   " calculated ground control point")
                    file_riw.write("\n")
                    file_riw.write("name_point  res_x  res_y  res_z\n")
                    for data in self.res_image_world:
                        file_riw.write(f"{data[0][0]}  {data[1][0]}  {data[1][1]}  {data[1][2]}\n")
                    file_riw.close()
            except FileNotFoundError as e:
                raise ValueError("The path doesn't exist !!!") from e

            with open(path_miw, "w", encoding="utf-8") as file_miw:
                file_miw.write("Control point statistics file.\n")
                file_miw.write("\n")
                file_miw.write("\n")
                file_miw.write("Statistics on residual function image to world.\n")
                file_miw.write("Name_stat: [stat_X, stat_Y, stat_Z]\n")
                self.write_stat(file_miw, self.stat_image_world)
                file_miw.close()

        if self.res_world_image:
            with open(path_rwi, "w", encoding="utf-8") as file_rwi:
                file_rwi.write("Control point statistics file.\n")
                file_rwi.write("\n")
                file_rwi.write("\n")
                file_rwi.write("Residual control points from terrain to image function.\n")
                file_rwi.write("residual = image's ground point - calculated image's ground point")
                file_rwi.write("\n")
                file_rwi.write("name_point  name_shot  res_column  res_line\n")
                for data in self.res_world_image:
                    file_rwi.write(f"{data[0][0]}  {data[0][1]}  {data[1][0]}  {data[1][1]}\n")
                file_rwi.close()

            with open(path_mwi, "w", encoding="utf-8") as file_mwi:
                file_mwi.write("Control point statistics file.\n")
                file_mwi.write("\n")
                file_mwi.write("\n")
                file_mwi.write("Statistics on residual function world to image.\n")
                file_mwi.write("Name_stat: [stat_column, stat_line]\n")
                self.write_stat(file_mwi, self.stat_world_image)
                file_mwi.close()

    def write_stat(self, file: io.TextIOWrapper, data: dict) -> None:
        """
        Write dictionary data statistics of the stat_list function on a file.

        Args:
            file (io.TextIOWrapper): File to write on.
            data (dict): Dictionary on data statistics to stat_list.
        """
        file.write(f"Minimum_arithmetic: {data['Min_arith']['val']}\n")
        file.write(f"Associated_point: {data['Min_arith']['data']}\n")
        file.write(f"Minimum_absolute: {data['Min_abs']['val']}\n")
        file.write(f"Associated_point: {data['Min_abs']['data']}\n")
        file.write(f"Maximum_arithmetic: {data['Max_arith']['val']}\n")
        file.write(f"Associated_point: {data['Max_arith']['data']}\n")
        file.write(f"Maximum_absolute: {data['Max_abs']['val']}\n")
        file.write(f"Associated_point: {data['Max_abs']['data']}\n")
        file.write(f"Median_arithmetic: {data['Median_arith']}\n")
        file.write(f"Median_absolute: {data['Median_abs']}\n")
        file.write(f"Mean_arithmetic: {data['Mean_arith']}\n")
        file.write(f"Mean_absolute: {data['Mean_abs']}\n")
        file.write(f"Variance_arithmetic: {data['Var_arith']}\n")
        file.write(f"Variance_absolute: {data['Var_abs']}\n")
        file.write(f"Sigma_arithmetic: {data['Sigma_arith']}\n")
        file.write(f"Sigma_absolute: {data['Sigma_abs']}\n")



"""
Camera data class module.
"""
from dataclasses import dataclass, field


@dataclass
class Camera:
    """
    Camera class definition.

    Args:
        name_camera (str): Name of the camera.
        ppax (float): Center of distortion in x.
        ppay (float): Center of distortion in y.
        focal (float): Focal of the camera.
        width (int): Width of the image in pixel.
        height (int): Height of the image in pixel.
    """
    name_camera: str
    ppax: float = field(default=None)
    ppay: float = field(default=None)
    focal: float = field(default=None)
    width: int = field(default=None)
    height: int = field(default=None)



"""
Photogrammetry worksite to writing in opk.
"""
import os
from pathlib import Path, PureWindowsPath
import numpy as np
from borea.worksite.worksite import Worksite
from borea.utils.check.check_args_opk import check_header_file


def write(name_opk: str, path_opk: str, args: dict, work: Worksite) -> None:
    """
    Write function, to save a photogrammetric site in .opk format.

    Args:
        name_opk (str): Name of the file writing.
        path_opk (str): Path of folder to registration file .opk.
        args (dict): Information for writing an opk file.
                     keys:
                     "order_axe" (str): Order of rotation matrix axes,
                     "header" (list): List of column type file.
                     "unit_angle" (str): Unit of angle 'degree' or 'radian'.
                     "linear_alteration" (bool): True if data corrected by
                     linear alteration.
        work (Worksite): The site to be recorded.
    """
    path_opk = os.path.join(Path(PureWindowsPath(path_opk)), f"{name_opk}.opk")

    if args["header"]:
        header, type_z = check_header_file(args["header"])
    else:
        header = ['N', 'X', 'Y', 'Z', 'O', 'P', 'K', 'C']
        type_z = work.type_z_shot

    if "S" in header:
        raise ValueError("Letter S doesn't existe in writing header opk.")

    work.set_unit_output(type_z, args["unit_angle"], args["linear_alteration"], args["order_axe"])

    header_file = ""
    for i in header:
        header_file += i + "   "

    try:
        with open(path_opk, "w", encoding="utf-8") as file:
            file.write(header_file)
            file.write("\n")
            keys = np.sort(list(work.shots))
            line_writing = ""
            for k in keys:
                shot = work.shots[k]
                dict_letter = {"N": shot.name_shot,
                               "X": str(shot.pos_shot[0]),
                               "Y": str(shot.pos_shot[1]),
                               "Z": str(shot.pos_shot[2]),
                               "O": str(shot.ori_shot[0]),
                               "P": str(shot.ori_shot[1]),
                               "K": str(shot.ori_shot[2]),
                               "C": shot.name_cam}
                for i in range(8):
                    line_writing += dict_letter[header[i]]
                    if i != 7:
                        line_writing += "   "
                    else:
                        line_writing += "\n"

            file.write(line_writing)
        file.close()
    except FileNotFoundError as e:
        raise ValueError("The path doesn't exist !!!", e) from e



"""
World image transformation module for Worksite
"""
from dataclasses import dataclass
from borea.worksite.worksite import Worksite
from borea.transform_world_image.transform_shot.world_image_shot import WorldImageShot


@dataclass
class WorldImageWork:
    """
    Class to calculate world coordinate to image coordinate in worksite.

    Args:
        name (str): Name of the worksite.
    """
    work: Worksite

    def calculate_world_to_image(self, lcode: list) -> None:
        """
        Calculates the position of gcp3d which corresponds to the data code
        in the images they appear in.

        Args:
            lcode (list): gcp code.
        """
        if self.work.gcp3d and self.work.gcp2d:
            for name_gcp, gcp in self.work.gcp3d.items():
                if gcp.code in lcode or lcode == []:
                    try:
                        list_shots = self.work.gcp2d[name_gcp]
                        for name_shot in list_shots:
                            shot = self.work.shots[name_shot]
                            cam = self.work.cameras[shot.name_cam]
                            coor_img = WorldImageShot(shot,
                                                      cam).world_to_image(gcp.coor,
                                                                          self.work.type_z_data,
                                                                          self.work.type_z_shot)
                            self.work.shots[name_shot].gcp3d[name_gcp] = coor_img
                    except KeyError:
                        print(f"Warning: id point {name_gcp} is present "
                              "in gcp3d but not in gcp2d.")
                        continue



"""
Class module for Rpc
"""
import numpy as np
from borea.datastruct.shot import Shot
from borea.datastruct.camera import Camera
from borea.geodesy.proj_engine import ProjEngine
from borea.datastruct.dtm import Dtm
from borea.transform_world_image.transform_shot.image_world_shot import ImageWorldShot
from borea.transform_world_image.transform_shot.world_image_shot import WorldImageShot
from borea.utils.miscellaneous.miscellaneous import normalize


class Rpc:
    """
    A class for computing RPC geometries.

    .. note::
        A RPC is always describing a transformation from
        geographical coordinates to images coordinates.
    """
    def __init__(self) -> None:
        self.param_rpc = {}
        self.fact_rpc = None
        self.output_epsg = False

    @classmethod
    def from_shot(cls, shot: Shot, cam: Camera, param_rpc: dict, unit_data: dict) -> None:
        """
        Calculate RPC of the shot.

        Args:
            shot (Shot): Shot to calculate rpc.
            cam (Camera): Camera of the shot.
            param_rpc (dict): Dictionary of parameters for rpc calculation.
            key;
            "size_grid"; size of the grip to calcule rpc.
            "order"; order of the polynome of the rpc.
            "fact_rpc"; rpc factor for world coordinate when src is not WGS84.
            unit_data (dict): Dictionary of data unity.
            key;
            "unit_z_data"; Unity of z data.
            "unit_z_shot"; Unity of z shot.

        Returns:
            Rpc: The object rpc.
        """
        obj = cls()
        obj.fact_rpc = param_rpc["fact_rpc"]
        obj.output_epsg = param_rpc["epsg_output"]
        obj.param_rpc["ERR_BIAS"] = -1
        obj.param_rpc["ERR_RAND"] = -1

        grid_img, grid_world = obj.create_grid_rpc(shot, cam, unit_data, param_rpc["size_grid"])

        img_norm, world_norm = obj.normalize_data(grid_img, grid_world)

        coeffx = obj.least_square_rpc(img_norm[0], world_norm, param_rpc["order"])
        coeffy = obj.least_square_rpc(img_norm[1], world_norm, param_rpc["order"])

        # Save param
        obj.param_rpc["SAMP_NUM_COEFF"] = coeffx[0:20]  # X numerator
        obj.param_rpc["SAMP_DEN_COEFF"] = coeffx[20:]  # X denominator
        obj.param_rpc["LINE_NUM_COEFF"] = coeffy[0:20]  # Y numerator
        obj.param_rpc["LINE_DEN_COEFF"] = coeffy[20:]  # Y denominator
        return obj

    def create_grid_rpc(self, shot: Shot, cam: Camera,
                        unit_data: dict, size_grid: int = 100) -> tuple:
        """
        Creation of a grid on a size_grid buffer around the acquisition.

        Args:
            shot (Shot): The shot to create the grid.
            cam (Camera): The camera of the shot.
            unit_data (dict): Dictionary of data unity.
            key;
            "unit_z_data"; Unity of z data.
            "unit_z_shot"; Unity of z shot.
            size_grid (int): The size of the grid to add around the shot.

        Returns:
            tuple: Image coordinates [col, line] and world coordinates [X, Y, Z] of the grid.
        """
        corner_img = np.array([[0, cam.width, 0, cam.width], [0, cam.height, cam.height, 0]])
        pt_w = ImageWorldShot(shot, cam).image_to_world(corner_img,
                                                        unit_data["unit_z_data"],
                                                        unit_data["unit_z_shot"])
        pt_min = np.min(pt_w, axis=1)[0:2] - size_grid - 1
        pt_max = np.max(pt_w, axis=1)[0:2] + size_grid + 1
        x, y = np.mgrid[int(pt_min[0]):int(pt_max[0]) + size_grid:size_grid,
                        int(pt_min[1]):int(pt_max[1]) + size_grid:size_grid]
        x, y = x.ravel(), y.ravel()
        z = np.squeeze(Dtm().get_z_world(np.array([x, y])))
        pt_img = WorldImageShot(shot, cam).world_to_image(np.array([x, y, z]),
                                                          unit_data["unit_z_data"],
                                                          unit_data["unit_z_shot"])
        return pt_img, np.array([x, y, z])

    def normalize_data(self, grid_img: np.ndarray, grid_world: np.ndarray) -> tuple:
        """
        Normalize data grid to calculate Rpc.

        Args:
            grid_img (np.array): Image coordinates [c, l] of the grid.
            grid_world (np.array): World coordinates [x, y, z] of the grid.

        Returns:
            tuple: grid image normalize and grid world normalize.
        """
        if self.fact_rpc is None:
            if self.output_epsg:
                x_geog, y_geog, z_geog = ProjEngine().tf.proj_to_proj_out(grid_world[0],
                                                                          grid_world[1],
                                                                          grid_world[2])
            else:
                x_geog, y_geog, z_geog = ProjEngine().tf.carto_to_geog(grid_world[0],
                                                                       grid_world[1],
                                                                       grid_world[2])
        else:
            x_geog = grid_world[0]*self.fact_rpc
            y_geog = grid_world[1]*self.fact_rpc
            z_geog = grid_world[2]

        # Normalization of image coordinates
        col_n_o_s = normalize(grid_img[0])
        lin_n_o_s = normalize(grid_img[1])

        # Normalization of world coordinates
        x_n_o_s = normalize(x_geog)
        y_n_o_s = normalize(y_geog)
        z_n_o_s = normalize(z_geog)

        self.param_rpc["LINE_OFF"] = lin_n_o_s[1]
        self.param_rpc["SAMP_OFF"] = col_n_o_s[1]
        self.param_rpc["LAT_OFF"] = y_n_o_s[1]
        self.param_rpc["LONG_OFF"] = x_n_o_s[1]
        self.param_rpc["HEIGHT_OFF"] = z_n_o_s[1]
        self.param_rpc["LINE_SCALE"] = lin_n_o_s[2]
        self.param_rpc["SAMP_SCALE"] = col_n_o_s[2]
        self.param_rpc["LAT_SCALE"] = y_n_o_s[2]
        self.param_rpc["LONG_SCALE"] = x_n_o_s[2]
        self.param_rpc["HEIGHT_SCALE"] = z_n_o_s[2]

        coor_img = np.array([col_n_o_s[0], lin_n_o_s[0]])
        coor_world = np.array([x_n_o_s[0], y_n_o_s[0], z_n_o_s[0]])

        return coor_img, coor_world

    def least_square_rpc(self, img_norm: np.ndarray, world_norm: np.ndarray,
                         polynomial_degree: int) -> np.ndarray:
        """
        Calcule Rpc of the shot by least square methode for one image coordinate.

        Args:
            img_norm (np.array): Normalize coordinates column or line of point in image.
            world_norm (np.array): Normalize coordinate of world grid.
            polynomial_degree (int): Degree of the polynomial of the rpc (1, 2, 3).

        Returns:
            np.array: Rpc coefficients.
        """
        mat_obs = self.setup_matrix_obs_rpc(img_norm, world_norm, polynomial_degree)
        x = np.linalg.lstsq(mat_obs, img_norm, rcond=None)[0]

        coef_rpc = np.zeros(40)
        if polynomial_degree == 1:
            coef_rpc[0:4] = x[0:4]
            coef_rpc[20] = 1
            coef_rpc[21:24] = x[4:7]

        elif polynomial_degree == 2:
            coef_rpc[0:10] = x[0:10]
            coef_rpc[20] = 1
            coef_rpc[21:30] = x[10:19]

        elif polynomial_degree == 3:
            coef_rpc[0:20] = x[0:20]
            coef_rpc[20] = 1
            coef_rpc[21:40] = x[20:39]

        return coef_rpc

    def setup_matrix_obs_rpc(self, img_norm: np.ndarray, world_norm: np.ndarray,
                             polynomial_degree: int) -> np.ndarray:
        """
        Setup observation matrix of the rpc for least square methode.

        Args:
            img_norm (np.array): Normalize coordinates column or line of point in image.
            world_norm (np.array): Normalize coordinate of world grid.
            polynomial_degree (int): Degree of the polynomial of the rpc (1, 2, 3).

        Returns:
            np.array: Observation matrix of the Rpc.
        """
        mat_a1 = []
        mat_a2 = []
        ligne_a = np.ones((len(img_norm)))
        mat_a1.append(ligne_a)

        if polynomial_degree in [1, 2, 3]:
            mat_a1.append(ligne_a * world_norm[0])
            mat_a1.append(ligne_a * world_norm[1])
            mat_a1.append(ligne_a * world_norm[2])

            mat_a2.append(ligne_a * -img_norm * world_norm[0])
            mat_a2.append(ligne_a * -img_norm * world_norm[1])
            mat_a2.append(ligne_a * -img_norm * world_norm[2])
        else:
            raise ValueError("The degree of the polynomial must be 1, 2 or 3")

        if polynomial_degree in [2, 3]:
            mat_a1.append(ligne_a * world_norm[0] * world_norm[1])
            mat_a1.append(ligne_a * world_norm[0] * world_norm[2])
            mat_a1.append(ligne_a * world_norm[1] * world_norm[2])
            mat_a1.append(ligne_a * world_norm[0] * world_norm[0])
            mat_a1.append(ligne_a * world_norm[1] * world_norm[1])
            mat_a1.append(ligne_a * world_norm[2] * world_norm[2])

            mat_a2.append(ligne_a * -img_norm * world_norm[0] * world_norm[1])
            mat_a2.append(ligne_a * -img_norm * world_norm[0] * world_norm[2])
            mat_a2.append(ligne_a * -img_norm * world_norm[1] * world_norm[2])
            mat_a2.append(ligne_a * -img_norm * world_norm[0] * world_norm[0])
            mat_a2.append(ligne_a * -img_norm * world_norm[1] * world_norm[1])
            mat_a2.append(ligne_a * -img_norm * world_norm[2] * world_norm[2])

        if polynomial_degree == 3:
            mat_a1.append(ligne_a * world_norm[0] * world_norm[1] * world_norm[2])
            mat_a1.append(ligne_a * world_norm[0] * world_norm[0] * world_norm[0])
            mat_a1.append(ligne_a * world_norm[0] * world_norm[1] * world_norm[1])
            mat_a1.append(ligne_a * world_norm[0] * world_norm[2] * world_norm[2])
            mat_a1.append(ligne_a * world_norm[0] * world_norm[0] * world_norm[1])
            mat_a1.append(ligne_a * world_norm[1] * world_norm[1] * world_norm[1])
            mat_a1.append(ligne_a * world_norm[1] * world_norm[2] * world_norm[2])
            mat_a1.append(ligne_a * world_norm[0] * world_norm[0] * world_norm[2])
            mat_a1.append(ligne_a * world_norm[1] * world_norm[1] * world_norm[2])
            mat_a1.append(ligne_a * world_norm[2] * world_norm[2] * world_norm[2])

            mat_a2.append(ligne_a * -img_norm * world_norm[0] * world_norm[1] * world_norm[2])
            mat_a2.append(ligne_a * -img_norm * world_norm[0] * world_norm[0] * world_norm[0])
            mat_a2.append(ligne_a * -img_norm * world_norm[0] * world_norm[1] * world_norm[1])
            mat_a2.append(ligne_a * -img_norm * world_norm[0] * world_norm[2] * world_norm[2])
            mat_a2.append(ligne_a * -img_norm * world_norm[0] * world_norm[0] * world_norm[1])
            mat_a2.append(ligne_a * -img_norm * world_norm[1] * world_norm[1] * world_norm[1])
            mat_a2.append(ligne_a * -img_norm * world_norm[1] * world_norm[2] * world_norm[2])
            mat_a2.append(ligne_a * -img_norm * world_norm[0] * world_norm[0] * world_norm[2])
            mat_a2.append(ligne_a * -img_norm * world_norm[1] * world_norm[1] * world_norm[2])
            mat_a2.append(ligne_a * -img_norm * world_norm[2] * world_norm[2] * world_norm[2])

        return np.concatenate((np.array(mat_a1), np.array(mat_a2)), axis=0).T



"""
Photogrammetry site file writer module.
"""
import importlib
from borea.worksite.worksite import Worksite


def manager_writer(writer: str, name: str, pathreturn: str, args: dict, work: Worksite) -> None:
    """
    Photogrammetric site file writing function.

    Args:
        writer (str): Output file format.
        name (str): Name of the file.
        pathreturn (str): Path to save the project.
        args (dict): Other information for writing the file.
        work (Worksite): The site to be recorded.
    """
    try:
        my_module = importlib.import_module("borea.writer.writer_" + writer.lower())
        work = my_module.write(name, pathreturn, args, work)
    except ModuleNotFoundError as e:
        raise ValueError(f"{writer} file is not taken into account !!!") from e



"""
Photogrammetry worksite to writing in rpc.
"""
import os
from pathlib import Path, PureWindowsPath
from borea.worksite.worksite import Worksite
from borea.geodesy.proj_engine import ProjEngine
from borea.format.conl import Conl
from borea.geodesy.projectionlist.search_proj import search_info


def write(name: str, folder_con: str, param_con: dict, work: Worksite) -> None:
    """
    Converte Worksite in Conical class and save it in CON.

    Args:
        name (str): Name of file begin.
        folder_con (str): Path of folder to registration file .CON.
        param_con (dict): None.
        work (Worksite): The site to be recorded.
    """
    _, _ = name, param_con

    if work.epsg_output:
        epsg_output = ProjEngine().epsg_output
    else:
        epsg_output = ProjEngine().epsg[0]

    work.set_unit_output(type_z="altitude", linear_alteration=True)
    geoview_proj = search_info("EPSG", str(epsg_output), "GEOVIEW")

    for name_shot, shot in work.shots.items():
        cam = work.cameras[shot.name_cam]
        path_conical = os.path.join(Path(PureWindowsPath(folder_con)), f"{name_shot}.CON")

        Conl(shot, cam, geoview_proj).save_conl(path_conical)



"""
Main to convert opk file to an light conical file (it's an xml file for GEOVIEW IGN France)
"""
# pylint: disable=import-error, wrong-import-position
import argparse
import sys
import os

sys.path.append(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))
from borea.process.p_format.p_read_opk import args_reading_opk, process_args_read_opk  # noqa: E402
from borea.process.p_format.p_write_con import args_write_con, process_args_write_con  # noqa: E402


def opk_to_conl():
    """
    Converts an OPK file into an light conical (IGN format) file.
    """
    parser = argparse.ArgumentParser(description='Photogrammetric site conversion '
                                                 'opk to conical file.')
    # Args for implement opk to conl
    parser = args_reading_opk(parser)
    parser = args_write_con(parser)

    args = parser.parse_args()

    # Process to read data
    work = process_args_read_opk(args)
    # Process to write conl
    process_args_write_con(args, work)


if __name__ == "__main__":
    opk_to_conl()



"""
Workdata data class module.
"""
import numpy as np
from pyproj import CRS, exceptions
from borea.datastruct.shot import Shot
from borea.datastruct.camera import Camera
from borea.datastruct.gcp import GCP
from borea.geodesy.proj_engine import ProjEngine
from borea.datastruct.dtm import Dtm


# pylint: disable-next=too-many-instance-attributes
class Workdata:
    """
    Workdata class.
    """
    def __init__(self, name: str) -> None:
        """
        Class definition of Workdata.

        Args:
            name (str): Name of the worksite.
        """
        self.name = name
        self.shots = {}
        self.cameras = {}
        self.co_points = {}
        self.gcp2d = {}
        self.gcp3d = {}
        self.co_pts_world = {}
        self.gcp2d_in_world = {}
        self.type_z_data = None
        self.type_z_shot = None
        self.approxeucli = False
        self.epsg_output = False

    # pylint: disable-next=too-many-arguments too-many-positional-arguments
    def add_shot(self, name_shot: str, pos_shot: np.ndarray,
                 ori_shot: np.ndarray, name_cam: str,
                 unit_angle: str, linear_alteration: bool,
                 order_axe: str) -> None:
        """
        Add Shot to the attribut Shots.

        Args:
            name_shot (str): Name of the shot.
            pos_shot (np.array): Array of coordinate position [X, Y, Z].
            ori_shot (np.array): Array of orientation of the shot [Omega, Phi, Kappa].
            name_cam (str): Name of the camera.
            unit_angle (str): Unit of angle 'degrees', 'radian'.
            linear_alteration (bool): True if z shot is correct of linear alteration.
            order_axe (str): Order of rotation matrix axes,
        """
        self.shots[name_shot] = Shot(name_shot=name_shot,
                                     pos_shot=pos_shot,
                                     ori_shot=ori_shot,
                                     name_cam=name_cam,
                                     unit_angle=unit_angle,
                                     linear_alteration=linear_alteration,
                                     order_axe=order_axe)

    def set_proj(self, epsg: list, path_geoid: list = None, epsg_output: int = None) -> None:
        """
        Setup a projection system to the worksite.

        Args:
            epsg (list): Code epsg of the projection ex: [2154].
            path_geoid (str): List of GeoTIFF which represents the geoid in grid form.
            epsg_output (int): Code epsg of the output projection. If you want to change.
        """
        ProjEngine.clear()
        for idepsg in [*epsg, epsg_output]:
            if idepsg:
                try:  # Check if the epsg exist
                    _ = CRS.from_epsg(idepsg)
                    del _
                except exceptions.CRSError as e_info:
                    raise exceptions.CRSError(f"Your EPSG:{epsg}"
                                              " doesn't exist in pyproj.") from e_info
        if epsg_output:
            self.epsg_output = True

        ProjEngine().set_epsg(epsg, path_geoid, epsg_output)

    # pylint: disable-next=too-many-arguments too-many-positional-arguments
    def add_camera(self, name_camera: str, ppax: float, ppay: float,
                   focal: float, width: int, height: int) -> None:
        """
        Add data camera in the Worksite.

        Args:
            name_camera (str): Name of the camera.
            ppax (float): Center of distortion in x.
            ppay (float): Center of distortion in y.
            focal (float): Focal of the camera.
            width (int): Width of the image camera.
            height (int): Height of the image camera.
        """
        self.cameras[name_camera] = Camera(name_camera=name_camera,
                                           ppax=ppax,
                                           ppay=ppay,
                                           focal=focal,
                                           width=width,
                                           height=height)

    def add_co_point(self, name_point: str, name_shot: str, coor2d: np.ndarray) -> None:
        """
        Add linking point between acquisition in two part.
        One in self.co_points a dict with name_point the key and list of acquisition the result.
        And One in self.shot[name_shot].co_points a dict whit
        name_point the key and list of coordinate x (column) y (line) the result in list.

        Agrs:
            name_point (str): Name of the connecting point.
            name_shot (str): Name of the acquisition.
            coor2d (array): Pixel position in the shot [x, y] = [column, line]
        """
        if name_shot not in self.shots:
            raise ValueError(f"The shot {name_shot} doesn't exist in list of shots.")

        if name_point not in self.co_points:
            self.co_points[name_point] = []

        if name_point not in self.shots[name_shot].co_points:
            self.shots[name_shot].co_points[name_point] = coor2d
        else:
            print("\n :--------------------------:")
            print("Warning : connecting point duplicate.")
            print(f"The point {name_point} already exists in the shot {name_shot}.")
            print("Keep first point with coordinates " +
                  f"{self.shots[name_shot].co_points[name_point]}.")
            print(":--------------------------:")

        self.co_points[name_point].append(name_shot)

    def add_gcp2d(self, name_point: str, name_shot: str, coor2d: np.ndarray) -> None:
        """
        Add linking point between acquisition in two part.
        One in self.gcp2d a dict with name_point the key
        and list of acquisition the result.
        And One in self.shot[name_shot].gcp2d a dict whit
        name_point the key and list of coordinate x (column) y (line) the result in list.

        Agrs:
            name_point (str): Name of the connecting point.
            name_shot (str): Name of the acquisition.
            coor2d (array): Pixel position in the shot [x, y] = [column, line]
        """
        try:
            self.shots[name_shot]
        except KeyError as e_info:
            raise ValueError(f"The shot {name_shot} doesn't exist in list of shots.") from e_info

        if name_point not in self.gcp2d:
            self.gcp2d[name_point] = []

        if name_point not in self.shots[name_shot].gcp2d:
            self.shots[name_shot].gcp2d[name_point] = coor2d
        else:
            print("\n :--------------------------:")
            print("Warning : connecting point duplicate.")
            print(f"The point {name_point} already exists in the shot {name_shot}.")
            print("Keep first point with coordinates " +
                  f"{self.shots[name_shot].gcp2d[name_point]}.")
            print(":--------------------------:")

        self.gcp2d[name_point].append(name_shot)

    def add_gcp3d(self, name_gcp: str, code_gcp: str, coor_gcp: np.ndarray) -> None:
        """
        Add GCP in the Worksite.

        Args:
            name_gcp (str): Name of the gcp.
            code_gcp (str): IGN code to differentiate between support points (1, 2, 3)
                            and control points (11, 12, 13).
                            1 means precision in Z, 2 in X and Y and 3 in X, Y, Z.
            coor_gcp (numpy.array): Array of ground coordinate [X, Y, Z].
        """
        self.gcp3d[name_gcp] = GCP(name_gcp, code_gcp, coor_gcp)

    def set_dtm(self, path_dtm: str, type_dtm: str) -> None:
        """
        Set class DtM to the worksite.

        Args:
            path_dtm (str): Path to the dtm.
            type (str): Type of the dtm "altitude" or "height".
        """
        if path_dtm:
            if type_dtm not in ["altitude", "height"]:
                raise ValueError(f"The dtm's type {type_dtm} isn't correct"
                                 " ('altitude' or 'height')")

            Dtm.clear()
            Dtm().set_dtm(path_dtm, type_dtm)

    def set_approx_eucli_proj(self, approx: bool) -> None:
        """
        Setup approxeucli in worksite.

        Args:
            apprx (bool): True if there are not projengine.
        """
        self.approxeucli = approx

    def set_type_z_shot(self, type_z_shot: str) -> None:
        """
        Setup type_z_shot in worksite.

        Args:
            type_z_shot (str): altitude or height.
        """
        self.type_z_shot = type_z_shot

    def set_type_z_data(self, type_z_data: str) -> None:
        """
        Setup type_z_data with one condition if is None.

        Args:
            type_z_data (str): altitude or height.
        """
        if type_z_data:
            self.type_z_data = type_z_data
        else:
            self.type_z_data = self.type_z_shot



"""
Image world transformation module for worksite
"""
from dataclasses import dataclass
from borea.worksite.worksite import Worksite
# pylint: disable-next=line-too-long
from borea.transform_world_image.transform_worksite.image_world_intersection import WorldIntersection  # noqa: E501
from borea.transform_world_image.transform_worksite.image_world_least_square import WorldLeastSquare


@dataclass
class ImageWorldWork:
    """
    Class to calculate image coordinate to world coordinate in worksite.

    Args:
        name (str): Name of the worksite.
    """
    work: Worksite

    def manage_image_world(self, type_point: str = "co_points", type_process: str = "inter",
                           control_type: list = None) -> None:
        """
        Process to calcule image coordinate to world coordinate.

        Args:
            type_point (str): "co_points" or "gcp2d"
                              depending on what you want to calculate.
            type_process (str): Type of process you want to use.
                                * "inter" by intersect bundle of point in each shot.
                                * "square" take all point and do methode least square.
            control_type (list): Type controle for gcp.
        """
        if type_point not in ["co_points", "gcp2d"]:
            raise ValueError(f"type_point {type_point} is incorrect,['co_points','gcp2d']")

        if type_process not in ["inter", "square"]:
            raise ValueError(f"type_process {type_process} is incorrect, "
                             "['inter','square']")

        if control_type is None or type_point == "co_points":
            control_type = []

        if type_process == "inter":
            WorldIntersection(self.work).calculate_image_world_by_intersection(type_point,
                                                                               control_type)

        if type_process == "square":
            WorldLeastSquare(self.work).compute_image_world_least_square(type_point, control_type)



"""
A module for manipulating an approximate euclidean projection.
"""
import numpy as np
from borea.geodesy.euclidean_proj import EuclideanProj


class ApproxEuclideanProj(EuclideanProj):
    """
     This class represents an approximate Euclidean projection system.

    Args:
        x_central (float): x coordinate of the central point of the Euclidean system.
        y_central (float): y coordinate of the central point of the Euclidean system.

    .. note::
        A Euclidean system is a local Euclidean reference system in which
        the collinear equation is valid.
    """
    def __init__(self, x_central: float, y_central: float) -> None:
        super().__init__(x_central, y_central)
        self.earth_raduis = 6366000

    def mat_to_mat_eucli(self, x: float, y: float, mat: np.ndarray) -> np.ndarray:
        """
        Transform the rotation matrix (World system) into rotation matrix (Euclidian systeme).

        Args:
            x (float): x coordinate of the point.
            y (float): y coordinate of the point.
            mat (np.array): Rotation matrix (World system).

        Returns:
            np.array: Euclidean rotation matrix.
        """
        _, _ = x, y
        return mat

    def mat_eucli_to_mat(self, x: float, y: float, mat_eucli: np.ndarray) -> np.ndarray:
        """
        Transform the rotation matrix (Euclidean system) into rotation matrix (World system).

        Args:
            x (float): x coordinate of the point.
            y (float): y coordinate of the point.
            mat_eucli (np.array): Rotation matrix (Euclidean system).

        Returns:
            np.array: Rotation matrix (World system).
        """
        _, _ = x, y
        return mat_eucli

    def world_to_eucli(self, coor_world: np.ndarray) -> np.ndarray:
        """
        Transform a point from the world coordinate reference system into
        the Euclidean coordinate reference system.

        Args:
            coor_world (np.array): World coordinate of the point [X, Y, Z].

        Returns:
            np.array: Point in the Euclidean coordinate reference system.
        """
        dx = coor_world[0] - self.pt_central[0]
        dy = coor_world[1] - self.pt_central[1]
        tt = (dx ** 2 + dy ** 2) / (4 * self.earth_raduis)
        cc = (self.earth_raduis + coor_world[2]) / (tt + self.earth_raduis)
        xe = dx * cc
        ye = dy * cc
        ze = ((self.earth_raduis - tt) * cc) - self.earth_raduis
        return np.array([xe, ye, ze])

    def eucli_to_world(self, coor_eucli: np.ndarray) -> np.ndarray:
        """
        Transform a point from the Euclidean coordinate reference system into
        the world coordinate reference system.

        Args:
            coor_eucli: Euclidean coordinate of the point [X, Y, Z].

        Returns:
            np.array: Point in the World coordinate reference system.
        """
        rz = self.earth_raduis + coor_eucli[2]
        rh = (rz**2 + coor_eucli[0]**2 + coor_eucli[1]**2)**0.5
        rz = (2*self.earth_raduis)/(rz + rh)
        x = self.pt_central[0] + (coor_eucli[0] * rz)
        y = self.pt_central[1] + (coor_eucli[1] * rz)
        z = rh - self.earth_raduis
        return np.array([x, y, z])



"""
Class to calcule world coordinate by least square methode.
"""
from dataclasses import dataclass
import numpy as np
import pandas as pd
from scipy.sparse import coo_matrix
from borea.worksite.worksite import Worksite
from borea.geodesy.local_euclidean_proj import LocalEuclideanProj
from borea.geodesy.approx_euclidean_proj import ApproxEuclideanProj
from borea.datastruct.dtm import Dtm
from borea.transform_world_image.transform_shot.image_world_shot import ImageWorldShot
from borea.transform_world_image.transform_shot.world_image_shot import WorldImageShot
from borea.utils.miscellaneous.sparse import invert_diag_sparse_matrix_3_3
from borea.utils.miscellaneous.param_bundle import set_param_bundle_diff


@dataclass
class WorldLeastSquare:
    """
    Class to calculate image coordinate to world coordinate in worksite by least square.

    Args:
        name (str): Name of the worksite.
    """
    work: Worksite

    def compute_image_world_least_square(self, type_point: str, control_type: list) -> None:
        """
        Calculates the mean of the result x y z of the point for each shot where it's visible.

        Args:
            type_point (str): "co_points" or "gcp2d"
                              depending on what you want to calculate.
            control_type (list): List of code gcp to take.
                                 To take all points or co_point, control_type = [].
        """
        # Creation of worksite's barycentre.
        bary = self.work.calculate_barycentre()
        if self.work.approxeucli:
            eucliproj = ApproxEuclideanProj(bary[0], bary[1])
        else:
            eucliproj = LocalEuclideanProj(bary[0], bary[1])

        # Retrieving point data from images.
        pd_mes = self.work.get_point_image_dataframe(type_point, control_type)
        pd_mes = pd_mes[pd_mes.duplicated(subset=['id_pt'], keep=False)]

        # Initialization of euclidean points.
        pd_pnt = self.init_eucli_points(pd_mes)

        # Do least square methode.
        pd_pnt = self.least_square_intersect(pd_mes, pd_pnt)

        # Transform euclidean point to world point.
        xw, yw, zw = eucliproj.eucli_to_world(np.array([pd_pnt["x"].to_numpy(),
                                                        pd_pnt["y"].to_numpy(),
                                                        pd_pnt["z"].to_numpy()]))
        pd_pnt["x"] = xw
        pd_pnt["y"] = yw
        pd_pnt["z"] = zw
        self.work.set_point_world_dataframe(pd_pnt, type_point)

    def init_eucli_points(self, pd_mes: pd.DataFrame) -> pd.DataFrame:
        """
        Initialization of ground points.

        Args:
            pd_mes (pd.Dataframe): Dataframe of image data, id_pt, id_img; column, line.

        Returns:
            pd.Dataframe: Dataframe of eucli data, id_pt, x, y, z.
        """
        pb_pnt_unique = pd_mes[~pd_mes['id_pt'].duplicated(keep='first')]
        group = pb_pnt_unique.groupby('id_img')
        frames = []

        for _, (id_shot, pd_mes_pnt) in enumerate(group):
            shot = self.work.shots[id_shot]
            cam = self.work.cameras[shot.name_cam]
            coor_img = np.array([pd_mes_pnt["column"].to_numpy(), pd_mes_pnt["line"].to_numpy()])
            z_world = np.squeeze(np.full(pd_mes_pnt.shape[0], Dtm().get_z_world(shot.pos_shot[:2])))

            # Calculation of world coordinate for all points in this shot
            coor_world = ImageWorldShot(shot, cam).image_z_to_world(coor_img,
                                                                    self.work.type_z_shot,
                                                                    z_world)

            # Transform of world coordinate to euclidean coordinate
            coor_eucli = shot.projeucli.world_to_eucli(coor_world)
            frames += [pd.DataFrame({"id_pt": pd_mes_pnt["id_pt"],
                                     "x": coor_eucli[0],
                                     "y": coor_eucli[1],
                                     "z": coor_eucli[2]})]

        pd_pnt = pd.concat(frames, ignore_index=True)
        pd_pnt["index_pnt"] = pd_pnt.index

        return pd_pnt

    def least_square_intersect(self, pd_mes: pd.DataFrame, pd_pnt: pd.DataFrame) -> pd.DataFrame:
        """
        Methode of least square to calcule world coordinate point.

        Args:
            pd_mes (pd.Dataframe): Dataframe of image data, id_pt, id_img; column, line.
            pd_pnt (pd.Dataframe): Dataframe of world data, id_pt, x, y, z.

        Returns:
            pd.Dataframe: Field data converged.
        """
        # Initialzsation number of observation
        nbr_obs = pd_mes.shape[0] * 2
        nbr_inc = pd_mes.nunique(axis=0)["id_pt"] * 3

        dx = 1.01
        nbr_iter = 0
        while not np.all(abs(dx) < 0.01) and nbr_iter < 5:
            # join data
            pd_mes_temp = pd_mes.join(pd_pnt.set_index('id_pt'), on="id_pt")
            pd_mes_temp = pd_mes_temp.sort_values(by=["index_pnt"]).reset_index(drop=True)
            pd_mes_temp["index_mes"] = pd_mes_temp.index

            mat_a, v_res = self.create_mat_a_and_vect_residu(pd_mes_temp, nbr_obs, nbr_inc)

            # Solving the system by inverting the block diagonal matrix
            dx = invert_diag_sparse_matrix_3_3(mat_a.T @ mat_a) @ mat_a.T @ v_res

            # Update World coordinate
            pd_pnt["x"] += dx[0::3]
            pd_pnt["y"] += dx[1::3]
            pd_pnt["z"] += dx[2::3]

            nbr_iter += 1

        return pd_pnt

    def create_mat_a_and_vect_residu(self, pd_mes_temp: pd.DataFrame,
                                     nbr_obs: int, nbr_inc: int) -> tuple:
        """
        Creation of matrix A and vector residu for the least square methode.

        Args:
            pd_mes_temp (pd.Dataframe): Dataframe of data.
            nbr_obs (int): Number of image observation.
            nbr_inc (int): Number of world observation.

        Returns:
            tuple: Matrix A and vector residu.
        """
        v_res = np.zeros(nbr_obs)
        coord_i, coord_j, data = [], [], []
        for id_shot, pd_data in pd_mes_temp.groupby('id_img'):
            shot = self.work.shots[id_shot]
            cam = self.work.cameras[shot.name_cam]

            pti = WorldImageShot(shot, cam).eucli_to_image(np.array([pd_data["x"].to_numpy(),
                                                                     pd_data["y"].to_numpy(),
                                                                     pd_data["z"].to_numpy()]),
                                                           self.work.type_z_data,
                                                           self.work.type_z_shot)

            _, vect_u, mat_v = set_param_bundle_diff(shot,
                                                     np.array([pd_data["x"].to_numpy(),
                                                               pd_data["y"].to_numpy(),
                                                               pd_data["z"].to_numpy()]))

            coord_i += [np.repeat(2 * pd_data['index_mes'].to_numpy(), 6) +
                        np.tile([0, 0, 0, 1, 1, 1], len(pd_data["x"].to_numpy()))]
            coord_j += [np.repeat(3 * pd_data['index_pnt'].to_numpy(), 6) +
                        np.tile([0, 1, 2, 0, 1, 2], len(pd_data["y"].to_numpy()))]

            data += [(np.tile(np.repeat(cam.focal / vect_u[2]**2, 2), (3, 1)).T *
                      mat_v @ shot.mat_rot_eucli).flatten()]

            v_res[2 * pd_data['index_mes'].to_numpy()] = pd_data["column"].to_numpy() - pti[0]
            v_res[2 * pd_data['index_mes'].to_numpy() + 1] = pd_data["line"].to_numpy() - pti[1]

        # Creation matrix
        coord_i = np.concatenate(coord_i)
        coord_j = np.concatenate(coord_j)
        data = np.concatenate(data)

        return coo_matrix((data, (coord_i, coord_j)), shape=(nbr_obs, nbr_inc)).tocsr(), v_res



"""
Args of parser for world coordinate of image point
"""
import argparse
import numpy as np
from borea.transform_world_image.transform_shot.image_world_shot import ImageWorldShot
from borea.worksite.worksite import Worksite


def args_add_pt2d(parser: argparse) -> argparse:
    """
    Args for 2D point

    Args:
        parser (argparse): Parser to add argument.

    Returns:
        argsparse: Parser with argument.
    """
    parser.add_argument('-p', '--point2d',
                        type=float, nargs=2,
                        help='Coordinate of the 2D point Column Line.')
    parser.add_argument('-d', '--type_z_data',
                        type=str, default=None, choices=['altitude', 'height', None],
                        help='Unit of the z of the data in output.')

    return parser


def process_image_world(args: argparse, work: Worksite) -> None:
    """
    Processing args to calculate world coordinate of the point.

    Args:
        args (argparse): Arg to apply on worksite (data).
        work (Worksite): Worksite to work on.
    """
    work.set_type_z_data(args.type_z_data)

    coor3d = ImageWorldShot(work.shots[args.name_shot],
                            work.cameras[list(work.cameras.keys())[0]]
                            ).image_to_world(np.array(args.point2d),
                                             work.type_z_data,
                                             work.type_z_shot)
    print(f"World coordinate of the point 2D {args.point2d} is : {np.round(coor3d, 2)}")
    print("Data summary:")
    print(f"* position shot {work.shots[args.name_shot].pos_shot}")
    print(f"* orientation shot {work.shots[args.name_shot].ori_shot}")



"""
A script for verification header str in manage reader.
"""
from borea.utils.check.check_header import check_head, check_h_z, get_type_z_and_header


def check_args_opk(args: dict) -> tuple:
    """
    Check args for reading an oopk.

    Args:
        args (dict): Information for reading an opk file.
                     keys:
                     "interval" (list): Interval of lines taken into account,
                     [i, j] if i or j is None = :. e.g. [1, None] = [1:].
                     "header" (list): List of column type file.
                     "unit_angle" (str): Unit of angle 'degrees' or 'radian'.
                     "linear_alteration" (bool): True if data corrected by linear alteration.

    Return:
        tuple: args, header and type of z shot.
    """
    if args["unit_angle"] not in ["degree", "radian"]:
        raise ValueError(f"Unit angles is incorrect {args['unit_angle']},"
                         "correct writing is degree or radian.")

    if args["interval"][0] is not None and args["interval"][0] > 0:
        args["interval"][0] -= 1
    if args["interval"][1] is not None:
        args["interval"][1] -= 1

    header, type_z = check_header_file(args["header"])

    return args, header, type_z


def check_header_file(header: list) -> tuple:
    """
    Check if the header of the file is good.

    Args:
        header (list): List of column type file.

    Returns:
        tuple: Header without type, type of z, type of angle.
    """
    list_letter = ['S', 'N', 'X', 'Y', 'Z', 'H', 'O', 'P', 'K', 'C']

    bad_head, ms_error_letter, head, symbol = check_head(header, list_letter)

    misss = set(list_letter[1:]) - symbol
    bad_head, ms_error_letter = check_h_z(bad_head, misss, ms_error_letter)

    ms_error = "Your header is not correct.\n"
    ms_error += ms_error_letter
    if bad_head:
        raise ValueError(ms_error)

    return get_type_z_and_header(head)



"""
Args of parser to read gcp 2D file.
"""
import argparse
from borea.worksite.worksite import Worksite
from borea.reader.reader_point import read_file_pt


def args_gcp2d(parser: argparse) -> argparse:
    """
    Args to read gcp 2D file.

    Args:
        parser (argparse): Parser to add argument.

    Returns:
        argsparse: Parser with argument.
    """
    parser.add_argument('-t', '--gcp2d',
                        type=str, default=None,
                        help='File path of ground control points in images.')
    parser.add_argument('-k', '--head_gcp2d',
                        type=str, default="PNXY",
                        help='Header of the file gcp2d.'
                        'e.g. PNXY'
                        'S: to ignore the column'
                        'P: name of point'
                        'N: name of shot'
                        'X: coordinate x (column) in the image'
                        'Y: coordinate y (line) in the image')
    return parser


def process_gcp2d(args, work: Worksite) -> Worksite:
    """
    Processing args with data.

    Args:
        args (argparse): Arg to apply on worksite (data).
        work (Worksite): Data.

    Returns:
        Worksite: data
    """
    # Reading ground point image
    if args.gcp2d is not None:
        read_file_pt(args.gcp2d, list(args.head_gcp2d.upper()), "gcp2d", work)
        print("Connecting point reading done.")
        count = 0
        for k in work.gcp2d.values():
            count += len(k)
        print(f"Number of gcp 2D in image: {len(work.gcp2d)}")
        print(f"Number of image with gcp 2D: {count}")

    return work



"""
Args of parser for image coordinate of ground point
"""
import argparse
import numpy as np
from borea.transform_world_image.transform_shot.world_image_shot import WorldImageShot
from borea.worksite.worksite import Worksite


def args_add_pt3d(parser: argparse) -> argparse:
    """
    Args for 3D point

    Args:
        parser (argparse): Parser to add argument.

    Returns:
        argsparse: Parser with argument.
    """
    parser.add_argument('-p', '--point3d',
                        type=float, nargs=3,
                        help='Coordinate of the 3D point X Y Z.')
    parser.add_argument('-d', '--type_z_data',
                        type=str, default=None, choices=['altitude', 'height', None],
                        help='Unit of the z of the data.')

    return parser


def process_world_image(args: argparse, work: Worksite) -> None:
    """
    Processing args to calculate image coordinate of the point.

    Args:
        args (argparse): Arg to apply on worksite (data).
        work (Worksite): Worksite to work on.
    """
    work.set_type_z_data(args.type_z_data)

    coor2d = WorldImageShot(work.shots[args.name_shot],
                            work.cameras[list(work.cameras.keys())[0]]
                            ).world_to_image(np.array(args.point3d),
                                             work.type_z_data,
                                             work.type_z_shot)
    print(f"Image coordinate of the point 3D {args.point3d} is : {np.round(coor2d, 2)}")
    print("Data summary:")
    print(f"* position shot {work.shots[args.name_shot].pos_shot}")
    print(f"* orientation shot {work.shots[args.name_shot].ori_shot}")



"""
Args of parser for adding one shot
"""
import argparse
import numpy as np
from borea.process.p_add_data.p_gen_param import args_general_param, process_args_gen_param
from borea.process.p_add_data.p_unit_shot import args_input_shot
from borea.worksite.worksite import Worksite


def args_add_shot(parser: argparse) -> argparse:
    """
    Args for adding one shot.

    Args:
        parser (argparse): Parser to add argument.

    Returns:
        argsparse: Parser with argument.
    """
    parser.add_argument('-n', '--name_shot',
                        type=str, default='Test',
                        help='Name of the shot.')
    parser.add_argument('-s', '--pos_shot',
                        type=float, nargs=3,
                        help='Position of the shot X Y Z.')
    parser.add_argument('-t', '--unit_z_shot',
                        type=str, choices=['altitude', 'height'],
                        help='Unit of the z of the shot.')
    parser.add_argument('-o', '--ori_shot',
                        type=float, nargs=3,
                        help='Orientation of the shot Omega Phi Kappa.')
    parser = args_input_shot(parser)
    parser = args_general_param(parser)
    return parser


def process_add_shot(args: argparse) -> Worksite:
    """
    Processing args with data.

    Args:
        args (argparse): Arg to apply on worksite (data)

    Returns:
        Worksite: data
    """
    work = Worksite("Calc_one_img")
    work = process_args_gen_param(args, work)

    work.add_shot(args.name_shot,
                  np.array(args.pos_shot),
                  np.array(args.ori_shot),
                  work.cameras[list(work.cameras.keys())[0]].name_camera,
                  args.unit_angle,
                  args.linear_alteration,
                  args.order_axe_input)

    work.set_param_shot(args.approx_system)

    work.type_z_shot = args.unit_z_shot

    return work



"""
Args of parser for reading opk file
"""
import argparse
from borea.process.p_add_data.p_gen_param import args_general_param, process_args_gen_param
from borea.process.p_add_data.p_unit_shot import args_input_shot
from borea.worksite.worksite import Worksite
from borea.reader.orientation.manage_reader import reader_orientation


def args_reading_opk(parser: argparse) -> argparse:
    """
    Args for reading opk file.

    Args:
        parser (argparse): Parser to add argument.

    Returns:
        argsparse: Parser with argument.
    """
    # pylint: disable=duplicate-code
    parser.add_argument('-r', '--filepath',
                        type=str, help='File path of the workfile.')
    parser.add_argument('-i', '--header',
                        type=str, default="NXYZOPKC",
                        help='Type of each column in the site file.'
                        'e.g. NXYZOPKC'
                        'S: to ignore the column'
                        'N: name of shot'
                        'X: coordinate x of the shot position'
                        'Y: coordinate y of the shot position'
                        'Z: coordinate z of the shot position in altitude'
                        'H: coordinate z of the shot position in height'
                        'O: omega rotation angle'
                        'P: phi rotation angle'
                        'K: kappa rotation angle'
                        'C: name of the camera')
    parser.add_argument('-f', '--first_line',
                        type=int, default=1,
                        help='Line number to start file playback. First line in the file is 1.'
                             ' Does not take file header into account.')
    parser.add_argument('-z', '--last_line',
                        type=int, default=None,
                        help='Line number to end file playback.'
                             ' If not set, all lines below -f will be read.')
    parser = args_input_shot(parser)
    parser = args_general_param(parser)
    return parser


def process_args_read_opk(args: argparse) -> Worksite:
    """
    Processing args with data.

    Args:
        args (argparse): Arg to apply on worksite (data)

    Returns:
        Worksite: data
    """
    # Readind data
    if args.filepath is not None:
        if args.header is not None:
            work = reader_orientation(args.filepath, {"order_axe": args.order_axe_input.lower(),
                                                      "interval": [args.first_line, args.last_line],
                                                      "header": list(args.header.upper()),
                                                      "unit_angle": args.unit_angle,
                                                      "linear_alteration": args.linear_alteration})
            print("Orientation file reading done.")
            print(f"Number of image: {len(work.shots)}")
        else:
            raise ValueError("The header file is missing -i.")
    else:
        raise ValueError("The access road to the photogrammetric site is missing -r.")

    work = process_args_gen_param(args, work)

    return work



"""
Args of parser for reading projection parameters
"""
import argparse
import pyproj
from borea.worksite.worksite import Worksite


def args_proj_param(parser: argparse) -> argparse:
    """
    Args for adding projection parameter.

    Args:
        parser (argparse): Parser to add argument.

    Returns:
        argsparse: Parser with argument.
    """
    parser.add_argument('-e', '--epsg',
                        type=int, default=None,
                        help='EPSG codifier number of the reference system used e.g. "2154".')
    parser.add_argument('-y', '--pathgeoid',
                        type=str, nargs='*', default=None,
                        help='Path to the pyproj GeoTIFF of the geoid e.g../test/data/geoid.tif'
                             f' or they must be in {pyproj.datadir.get_data_dir()} and just'
                             ' need name of file e.g. geoid.tif.')
    parser.add_argument('--geog', '--epsg_geographic',
                        type=int, default=None,
                        help='EPSG codifier number of the reference geographic system.')
    parser.add_argument('--geoc', '--epsg_geocentric',
                        type=int, default=None,
                        help='EPSG codifier number of the reference geocentric system.')
    parser.add_argument('--oe', '--epsg_output',
                        type=int, default=None,
                        help="Code epsg of output Data")
    return parser


def process_args_proj_param(args: argparse, work: Worksite) -> Worksite:
    """
    Processing args with data.

    Args:
        args (argparse): Arg to apply on worksite (data).
        work (Worksite): Worksite to work on.

    Returns:
        Worksite: data
    """
    # Add a projection to the worksite
    if args.epsg is not None:
        work.set_proj([args.epsg, args.geog, args.geoc], args.pathgeoid, args.oe)
        print(f"Projection set-up with EPSG:{args.epsg}.")
    else:
        print("There is no given projection.")

    return work



"""
Research data in ProjectionList.txt.
"""
import os


PATH_FILE = os.path.join(os.path.dirname(__file__), "ProjectionList.txt")


def convert_line_list(line: str) -> list:
    """
    Convert type of all element in list in str.

    Args:
        list_data(str): One line of file.

    Returns:
        list: List with str object.
    """
    return [x[1:-1] for x in line[1:-2].split(',')]


def search_info(type_input: str, data_input: str, type_output: str) -> str:
    """
    Research data in ProjectionList.txt.
    header: ["SD","BDORTHO","EPSG","GEOVIEW","MAPINFO","PROJ4","TA","TOPAERO","WKT"]

    Args:
        type_input (str): type of data input to refer.
        data_input (str): data to refer.
        type_output (str): type of data output to research.

    Returns:
        str: data output.
    """
    try:
        with open(PATH_FILE, 'r', encoding="utf-8") as file_proj:
            list_data = file_proj.readlines()
            header = convert_line_list(list_data[0])
            if type_input.upper() not in header and type_output.upper() not in header:
                raise ValueError(f"{type_input} or {type_output} not in header: {header}")

            id_input = header.index(type_input)
            id_output = header.index(type_output)

            for info_proj in list_data[1:]:
                data = convert_line_list(info_proj)
                if data[id_input] == "null" or data[id_output] == "null":
                    continue
                if data[id_input] == data_input:
                    data_research = data[id_output]
                    break

        file_proj.close()
    except FileNotFoundError as e:
        raise FileNotFoundError(f"The path {PATH_FILE} is incorrect !!!") from e

    if data_research == "null":
        return None
    return data_research



"""
Main to calculate image coodinate of the ground point.
"""
# pylint: disable=import-error, wrong-import-position
import argparse
import sys
import os

sys.path.append(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))
from borea.process.p_add_data.p_add_shot import args_add_shot, process_add_shot  # noqa: E402
from borea.process.p_add_data.p_pt2d import args_add_pt2d, process_image_world  # noqa: E402


def pt_image_to_world():
    """
    Converts a image point into an terrain point.
    """
    parser = argparse.ArgumentParser(description='Calculate image coodinate of the ground point.')
    # Args for implement pt image to world
    parser = args_add_shot(parser)
    parser = args_add_pt2d(parser)

    args = parser.parse_args()

    # Process to read data
    work = process_add_shot(args)
    # Process to transform image to world
    process_image_world(args, work)


if __name__ == "__main__":
    pt_image_to_world()



"""
Args of parser to control file
"""
import argparse
from borea.worksite.worksite import Worksite
from borea.process.p_add_data.p_file_gcp2d import args_gcp2d, process_gcp2d
from borea.process.p_add_data.p_file_gcp3d import args_gcp3d, process_gcp3d
from borea.stat.statistics import Stat
from borea.transform_world_image.transform_worksite.world_image_work import WorldImageWork
from borea.transform_world_image.transform_worksite.image_world_work import ImageWorldWork


def args_control(parser: argparse) -> argparse:
    """
    Args to control photogrametrique file.

    Args:
        parser (argparse): Parser to add argument.

    Returns:
        argsparse: Parser with argument.
    """
    parser = args_gcp2d(parser)
    parser = args_gcp3d(parser)
    parser.add_argument('-d', '--control_type',
                        type=str, default=[], nargs='*',
                        help='Type of gcp to control.')
    parser.add_argument('-p', '--process',
                        type=str, default="inter", choices=["inter", "square"],
                        help="Type of process for the function image to world,"
                             "intersection or least_square")
    parser.add_argument('-w', '--pathreturn',
                        type=str, default='./',
                        help='Conversion path e.g. test/tmp/.')
    return parser


def process_args_control(args, work: Worksite) -> None:
    """
    Processing args with data.

    Args:
        args (argparse): Arg to apply on worksite (data).
        work (Worksite): Data.
    """
    # Read GCP 2D
    work = process_gcp2d(args, work)

    # Read GCP 3D
    work = process_gcp3d(args, work)

    # Calculate ground coordinate of conneting point by intersection
    print("Calculation of data image to world.")
    ImageWorldWork(work).manage_image_world(type_point="gcp2d",
                                            type_process=args.process,
                                            control_type=args.control_type)

    # Calculate image coordinate of GCP if they exist
    print("Calculation of data world to image.")
    WorldImageWork(work).calculate_world_to_image(args.control_type)

    # Statistics
    print("Make statistics.")
    stat = Stat(work, args.pathreturn, args.control_type)
    stat.main_stat_and_save()
    print("Statistics on control point, if there are,")
    print(f" in {args.pathreturn}Stat_module_{work.name}.txt .")



"""
A module for manipulating a local euclidean projection.
"""
import numpy as np
from scipy.spatial.transform import Rotation as R
from borea.geodesy.proj_engine import ProjEngine
from borea.geodesy.euclidean_proj import EuclideanProj
from borea.utils.check.check_array import check_array_transfo


class LocalEuclideanProj(EuclideanProj):
    """
    This class represents a Euclidean projection system.

    Args:
        x_central (float): x coordinate of the central point of the Euclidean system.
        y_central (float): y coordinate of the central point of the Euclidean system.

    .. note::
        A Euclidean system is a local Euclidean reference system in which
        the collinear equation is valid.
        The cartographic system parameters must be known.
    """
    def __init__(self, x_central: float, y_central: float) -> None:
        super().__init__(x_central, y_central)
        self.rot_to_euclidean_local = self.mat_rot_euclidean_local(x_central, y_central)

    def mat_rot_euclidean_local(self, x: float, y: float) -> np.ndarray:
        """
        Compute the transition matrix between the world system and
        the Euclidean system centred on a point.

        Args:
            x (float): x coordinate of the central point of the Euclidean system.
            y (float): y coordinate of the central point of the Euclidean system.

        Returns:
            np.array: Transition matrix.
        """
        lon, lat = ProjEngine().tf.carto_to_geog(x, y)
        gamma = ProjEngine().get_meridian_convergence(x, y)

        # Matrix for switching to local cartesian coordinates
        rot_ecef_to_eucli = R.from_euler("ZYZ", np.array([lon, 90 - lat, 90 + gamma]),
                                         degrees=True).inv()
        return rot_ecef_to_eucli.as_matrix()

    def mat_to_mat_eucli(self, x: float, y: float, mat: np.ndarray) -> np.ndarray:
        """
        Transform the rotation matrix (World system) into rotation matrix (Euclidian systeme).

        Args:
            x (float): x coordinate of the point.
            y (float): y coordinate of the point.
            mat (np.array): Rotation matrix (World system).

        Returns:
            np.array: Euclidean rotation matrix.
        """
        # We are in the projection system, we pass into the local tangeant system
        matecef_to_rtl = self.mat_rot_euclidean_local(x, y)
        mat_eucli = mat @ matecef_to_rtl @ self.rot_to_euclidean_local.T
        return mat_eucli

    def mat_eucli_to_mat(self, x: float, y: float, mat_eucli: np.ndarray) -> np.ndarray:
        """
        Transform the rotation matrix (Euclidean system) into rotation matrix (World system).

        Args:
            x (float): x coordinate of the point.
            y (float): y coordinate of the point.
            mat_eucli (np.array): Rotation matrix (Euclidean system).

        Returns:
            np.array: Rotation matrix (World system).
        """
        # We're in the local tangeant frame, now we're in the projection frame
        matecef_to_rtl = self.mat_rot_euclidean_local(x, y)
        mat = mat_eucli @ self.rot_to_euclidean_local @ matecef_to_rtl.T
        return mat

    def world_to_eucli(self, coor: np.ndarray) -> np.ndarray:
        """
        Transform a point from the world coordinate reference system into
        the Euclidean coordinate reference system.

        Args:
            coor (np.ndarray): Coordinate [X, Y, Z].

        Returns:
            np.array: x, y, z in the Euclidean coordinate reference system.
        """
        coor = check_array_transfo(coor[0], coor[1], coor[2])

        coor_geoc = np.array(ProjEngine().tf.carto_to_geoc(coor[0], coor[1], coor[2]))
        central_geoc = np.array(ProjEngine().tf.carto_to_geoc(self.pt_central[0],
                                                              self.pt_central[1],
                                                              self.pt_central[2]))
        dr = np.vstack([coor_geoc[0] - central_geoc[0],
                        coor_geoc[1] - central_geoc[1],
                        coor_geoc[2] - central_geoc[2]])
        point_eucli = np.squeeze((self.rot_to_euclidean_local @ dr) + np.array([self.pt_central]).T)
        return point_eucli

    def eucli_to_world(self, coor: np.ndarray) -> np.ndarray:
        """
        Transform a point from the Euclidean coordinate reference system into
        the world coordinate reference system.

        Args:
            coor (np.ndarray): Coordinate [X, Y, Z].

        Returns:
            np.array: x, y, z in the world coordinate reference system.
        """
        coor = np.squeeze(coor)

        central_geoc = np.array(ProjEngine().tf.carto_to_geoc(self.pt_central[0],
                                                              self.pt_central[1],
                                                              self.pt_central[2]))
        dr = np.vstack([coor[0] - self.pt_central[0],
                        coor[1] - self.pt_central[1],
                        coor[2] - self.pt_central[2]])
        point_geoc = np.squeeze((self.rot_to_euclidean_local.T @ dr) + np.array([central_geoc]).T)
        x_gc, y_gc, z_gc = check_array_transfo(point_geoc[0], point_geoc[1], point_geoc[2])
        tup = ProjEngine().tf.geoc_to_carto(x_gc, y_gc, z_gc)
        return np.array([tup[0], tup[1], tup[2]])



"""
Photogrammetry worksite to writing in rpc.
"""
import os
from pathlib import Path, PureWindowsPath
from borea.format.rpc import Rpc
from borea.worksite.worksite import Worksite
from borea.datastruct.dtm import Dtm


def write(name: str, folder_rpc: str, param_rpc: dict, work: Worksite) -> None:
    """
    Converte Worksite in RPC class and save it in txt.

    Args:
        name (str): Name of file begin.
        folder_rpc (str): Path of folder to registration file .txt.
        param_rpc (dict): Dictionary of parameters for rpc calculation.
        key;
        "size_grid"; size of the grip to calcule rpc.
        "order"; order of the polynome of the rpc.
        "fact_rpc"; rpc factor for world coordinate when src is not WGS84.
        "epsg_output"; code epsg for RPC.
        work (Worksite): The site to be recorded.
    """
    _ = name
    keys = ["ERR_BIAS", "ERR_RAND", "LINE_OFF", "SAMP_OFF",
            "LAT_OFF", "LONG_OFF", "HEIGHT_OFF", "LINE_SCALE",
            "SAMP_SCALE", "LAT_SCALE", "LONG_SCALE",
            "HEIGHT_SCALE"]

    param_rpc["epsg_output"] = work.epsg_output

    work.set_unit_output(type_z=Dtm().type_dtm, proj_output=False)

    for name_shot, shot in work.shots.items():
        cam = work.cameras[shot.name_cam]

        rpc = Rpc.from_shot(shot, cam, param_rpc,
                            {"unit_z_data": work.type_z_data, "unit_z_shot": work.type_z_shot})

        list_txt_rpc = [f"{key}: {rpc.param_rpc[key]}" for key in keys]

        for idx, val in enumerate(rpc.param_rpc["LINE_NUM_COEFF"]):
            list_txt_rpc += [f"LINE_NUM_COEFF_{idx + 1}: {val}"]

        for idx, val in enumerate(rpc.param_rpc["LINE_DEN_COEFF"]):
            list_txt_rpc += [f"LINE_DEN_COEFF_{idx + 1}: {val}"]

        for idx, val in enumerate(rpc.param_rpc["SAMP_NUM_COEFF"]):
            list_txt_rpc += [f"SAMP_NUM_COEFF_{idx + 1}: {val}"]

        for idx, val in enumerate(rpc.param_rpc["SAMP_DEN_COEFF"]):
            list_txt_rpc += [f"SAMP_DEN_COEFF_{idx + 1}: {val}"]

        path_rpc = os.path.join(Path(PureWindowsPath(folder_rpc)),
                                f"{name_shot}_RPC.TXT")
        Path(path_rpc).write_text("\n".join(list_txt_rpc), encoding="UTF-8")



"""
Main to calculate 6 externals parameters of shots and save in opk file.
"""
# pylint: disable=import-error, wrong-import-position, line-too-long
import argparse
import sys
import os

sys.path.append(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))
from borea.process.p_func.p_spaceresection import args_space_resection, process_space_resection  # noqa: E402, E501
from borea.process.p_format.p_write_opk import args_writing_opk, process_args_write_opk  # noqa: E402, E501


def spaceresection_opk():
    """
    Produces an OPK file from image link points and their field coordinates
    coordinates (calculates the 6 external parameters for each image with a point)
    """
    parser = argparse.ArgumentParser(description='Space resection of point file image and world'
                                                 ' to calculate 6 externals parameters of shots')
    # Args for implement space resection( opk
    parser = args_space_resection(parser)
    parser = args_writing_opk(parser)

    args = parser.parse_args()

    # Process to read data
    work = process_space_resection(args)
    # Process to space resection
    process_args_write_opk(args, work)


if __name__ == "__main__":
    spaceresection_opk()



"""
A module for manipulating a digital elevation model.
"""
from pathlib import Path, PureWindowsPath
import numpy as np
from osgeo import gdal
from scipy import ndimage
from borea.transform_world_image.transform_dtm.world_image_dtm import WorldImageDtm
from borea.utils.singleton.singleton import Singleton
gdal.UseExceptions()


# pylint: disable-next=too-many-instance-attributes
class Dtm(WorldImageDtm, metaclass=Singleton):
    """
    Represents a digital elevation model.

    Args:
        path_dtm (str): Path to the dtm file. If None Dtm.get() always returns  0.
        order (int): The method of interpolation to perform.
                    0 : nearest, 1 : slinear, 3 : cubic, 5 : quintic.
        cval (int): Value to fill past edges of dtm.
                    If None raise an error for any point outside the dtm.
        keep_in_memory (bool): Store all data in memory.

    .. note::
        All gdal formats are supported. The file must contain georeferencing information.
    """
    def __init__(self):
        self.path_dtm = None
        self.type_dtm = None
        self.order = 1
        self.keep_in_memory = False
        self.cval = np.nan
        self.img = None
        self.rb = None
        self.nodata = None
        self.dtm_array = None
        WorldImageDtm.__init__(self, None)

    def set_dtm(self, path_dtm: str, type_dtm: str) -> None:
        """
        Set the dtm path for reading dtm.

        Args:
            path_dtm (str): Path to the dtm file.
            type (str): Type of dtm "a" altitude, "h" height.
        """
        if path_dtm:
            gdal.AllRegister()
            self.type_dtm = type_dtm
            self.path_dtm = Path(PureWindowsPath(path_dtm))
            self.img = gdal.Open(self.path_dtm.as_posix())
            self.rb = self.img.GetRasterBand(1)
            self.nodata = self.rb.GetNoDataValue()
            if self.keep_in_memory:
                self.dtm_array = self.rb.ReadAsArray()
            WorldImageDtm.__init__(self, self.img.GetGeoTransform())
        else:
            self.path_dtm = path_dtm

    def set_order(self, order: int) -> None:
        """
        The method of interpolation to perform.
        0:nearest 1:bilinear 3:cubic 5:quintic.

        Args:
            order (int): The method of interpolation to perform.
        """
        self.order = order

    def set_cval(self, cval: int) -> None:
        """
        Value to fill past edges of dtm or nodata.
        If None raise an error for any point outside the dtm or Nodata.

        Args
            cval (int): Value to fill past edges of dtm or nodata.
        """
        self.cval = cval if cval else np.nan

    def set_keep_memory(self, keep_memory: bool) -> None:
        """
        Store all image in memory.

        Args:
            keep_in_memory (bool): Store all image in memory.
        """
        self.keep_in_memory = keep_memory

    def get_z_world(self, coor_2d: np.ndarray) -> np.ndarray:
        """
        Extract value in the Dtm.

        Args:
            coor_2d (np.array): World coordinate 2D [X, Y].

        Returns:
            np.array: z value.
        """
        if self.path_dtm is None:
            return np.zeros_like(coor_2d[0])
        col, line = self.world_to_image(coor_2d)
        if not self.keep_in_memory:
            z = []
            if isinstance(col, float):
                col, line = [col], [line]
            for colt, linet in zip(col, line):
                try:
                    z += [self.rb.ReadAsArray(colt, linet, 1, 1, resample_alg=self.order)]
                except RuntimeError:
                    z += [self.cval]
        else:
            z = ndimage.map_coordinates(self.dtm_array, np.vstack([line, col]),
                                        order=self.order, mode="constant", cval=self.cval)
        if np.any(np.isnan(z)):
            raise IndexError(f"Out dtm {coor_2d[0]} {coor_2d[1]}")

        return np.round(np.array(z, dtype=float), 3)



"""
Args of parser for writing conical file
"""
import argparse
from borea.worksite.worksite import Worksite
from borea.writer.manage_writer import manager_writer


def args_write_con(parser: argparse) -> argparse:
    """
    Args for writing conical file of GEOVIEW IGN.

    Args:
        parser (argparse): Parser to add argument.

    Returns:
        argsparse: Parser with argument.
    """
    parser.add_argument('-w', '--pathreturn',
                        type=str, default='./',
                        help='Conversion path e.g. test/tmp/.')
    return parser


def process_args_write_con(args: argparse, work: Worksite) -> None:
    """
    Processing args with data.

    Args:
        args (argparse): Arg to apply on worksite (data).
        work (Worksite): Data.
    """
    # Writing data
    print("Writing Conical file.")
    manager_writer("con", None, args.pathreturn, None, work)
    print(f"File written in folder {args.pathreturn} in .CON format.")



"""
Module for class ProjEngine, transform geodesy
"""
import pyproj
import numpy as np
import pandas as pd


# pylint: disable=too-many-instance-attributes
class TransformGeodesy():
    """
    This class provides functions to tranform coordinate system.
    Class parent of ProjEngine.
    Implemented by ProjEngine.

    Args:
        geoid (list): List of geoid to use.
        epsg (list): Code epsg of the projection ex: [2154].
        epsg_output (int): Code epsg of the output projection. If you want to change.
    """
    def __init__(self, epsg: list, geoid: list, epsg_output: int) -> None:
        self.epsg = epsg
        self.epsg_output = epsg_output
        self.crs = pyproj.CRS.from_epsg(epsg[0])
        self.geoid = geoid
        self._carto_to_geoc = None
        self._geoc_to_carto = None
        self._carto_to_geog = None
        self._geog_to_carto = None
        self._geog_to_geoid = None
        self._geoid_to_geog = None
        self._proj_to_proj_out = None

    @property
    def carto_to_geog(self) -> pyproj.Transformer:
        """
        Returns the transformation or instantiates it before returning it.

        Returns:
            pyproj.Transformer : carto_to_geog
        """
        if not self._carto_to_geog:
            try:
                crs_geog = pyproj.CRS.from_epsg(self.epsg[1])
            except (IndexError, pyproj.exceptions.CRSError):
                crs_geog = pyproj.crs.GeographicCRS(name=self.crs.name, datum=self.crs.datum.name)

            self._carto_to_geog = pyproj.Transformer.from_crs(self.crs, crs_geog).transform

        return self._carto_to_geog

    @property
    def geog_to_carto(self) -> pyproj.Transformer:
        """
        Returns the transformation or instantiates it before returning it.

        Returns:
            pyproj.Transformer : geog_to_carto
        """
        if not self._geog_to_carto:
            try:
                crs_geog = pyproj.CRS.from_epsg(self.epsg[1])
            except (IndexError, pyproj.exceptions.CRSError):
                crs_geog = pyproj.crs.GeographicCRS(name=self.crs.name, datum=self.crs.datum.name)

            self._geog_to_carto = pyproj.Transformer.from_crs(crs_geog, self.crs).transform

        return self._geog_to_carto

    @property
    def carto_to_geoc(self) -> pyproj.Transformer:
        """
        Returns the transformation or instantiates it before returning it.

        Returns:
            pyproj.Transformer : carto_to_geoc
        """
        if not self._carto_to_geoc:
            try:
                crs_geoc = pyproj.CRS.from_epsg(self.epsg[2])
            except (IndexError, pyproj.exceptions.CRSError):
                crs_geoc = pyproj.crs.GeocentricCRS(name=self.crs.name, datum=self.crs.datum.name)

            self._carto_to_geoc = pyproj.Transformer.from_crs(self.crs, crs_geoc).transform

        return self._carto_to_geoc

    @property
    def geoc_to_carto(self) -> pyproj.Transformer:
        """
        Returns the transformation or instantiates it before returning it.

        Returns:
            pyproj.Transformer : geoc_to_carto
        """
        if not self._geoc_to_carto:
            try:
                crs_geoc = pyproj.CRS.from_epsg(self.epsg[2])
            except (IndexError, pyproj.exceptions.CRSError):
                crs_geoc = pyproj.crs.GeocentricCRS(name=self.crs.name, datum=self.crs.datum.name)

            self._geoc_to_carto = pyproj.Transformer.from_crs(crs_geoc, self.crs).transform

        return self._geoc_to_carto

    @property
    def geoid_to_geog(self) -> pyproj.Transformer:
        """
        Returns the transformation or instantiates it before returning it.

        Returns:
            pyproj.Transformer : geoid_to_geog
        """
        if not self.geoid:
            raise ValueError("Mistake Geoid path")

        if not self._geoid_to_geog:
            try:
                # Transform geoide coordinates to geographic coordinates
                self._geoid_to_geog = pyproj.Transformer.from_pipeline("+proj=vgridshift "
                                                                       "+grids="
                                                                       f"{','.join(self.geoid)} "
                                                                       "+multiplier=1").transform
            except pyproj.exceptions.ProjError as e:
                raise pyproj.exceptions.ProjError(f"{self.geoid} The name or path "
                                                  "of geotif is incorrect or does not exist in "
                                                  f"{pyproj.datadir.get_data_dir()}!!!{e}") from e

        return self._geoid_to_geog

    @property
    def geog_to_geoid(self) -> pyproj.Transformer:
        """
        Returns the transformation or instantiates it before returning it.

        Returns:
            pyproj.Transformer : geog_to_geoid
        """
        if not self.geoid:
            raise ValueError("Mistake Geoid path")

        if not self._geog_to_geoid:
            try:
                # Transform geoide coordinates to geographic coordinates
                self._geog_to_geoid = pyproj.Transformer.from_pipeline("+proj=vgridshift "
                                                                       "+grids="
                                                                       f"{','.join(self.geoid)} "
                                                                       "+multiplier=-1").transform
            except pyproj.exceptions.ProjError as e:
                raise pyproj.exceptions.ProjError(f"{self.geoid} The name or path "
                                                  "of geotif is incorrect or does not exist in "
                                                  f"{pyproj.datadir.get_data_dir()}!!!{e}") from e

        return self._geog_to_geoid

    @property
    def proj_to_proj_out(self) -> pyproj.Transformer:
        """
        Create the pyproj Transformer from crs of worksite to crs geographic ask.

        Returns:
            pyproj.Transformer : carto_to_geog_out
        """
        if not self._proj_to_proj_out:
            crs_out = pyproj.CRS.from_epsg(self.epsg_output)
            self._proj_to_proj_out = pyproj.Transformer.from_crs(self.crs, crs_out).transform

        return self._proj_to_proj_out

    def tranform_height(self, coor: np.ndarray) -> float:
        """
        Converting z in altitude to z in height of point.

        Args:
            coor (np.array): [X, Y, Z] coordinate of the point.

        Returns:
            float: New height Z.
        """
        coor_geog = self.carto_to_geog(coor[0], coor[1], coor[2])
        try:
            coor_geog = self.geoid_to_geog(coor_geog[0],
                                           coor_geog[1],
                                           coor_geog[2])
            new_z = self.geog_to_carto(coor_geog[0],
                                       coor_geog[1],
                                       coor_geog[2])[2]
        except AttributeError as info:
            raise ValueError("The geoid has not been entered, "
                             "cannot transform z altitude to height.") from info

        if np.all(new_z == np.inf):
            raise ValueError("out geoid")
        return new_z

    def tranform_altitude(self, coor: np.ndarray) -> float:
        """
        Converting z in height to z in altitude of point.

        Args:
            coor (np.array): [X, Y, Z] coordinate of the point.

        Returns:
            float: New altitude Z.
        """
        coor_geog = self.carto_to_geog(coor[0], coor[1], coor[2])
        try:
            coor_geog = self.geog_to_geoid(coor_geog[0],
                                           coor_geog[1],
                                           coor_geog[2])
            new_z = self.geog_to_carto(coor_geog[0],
                                       coor_geog[1],
                                       coor_geog[2])[2]
        except AttributeError as info:
            raise ValueError("The geoid has not been entered, "
                             "cannot transform z height to altitude.") from info

        if np.all(new_z == np.inf):
            raise ValueError("out geoid")
        return new_z

    def transform_pt_proj(self, df_pt: pd.DataFrame, type_z_input: str = None,
                          type_z_output: str = None) -> pd.DataFrame:
        """
        Tranform the input projection to the output projection of points coordinates
        """
        if type_z_input and type_z_output:
            if type_z_input != type_z_output:
                if type_z_output == "altitude":
                    df_pt["z"] = self.tranform_altitude(np.array([df_pt['x'],
                                                                  df_pt['y'],
                                                                  df_pt['z']]))
                if type_z_output == "height":
                    df_pt["z"] = self.tranform_height(np.array([df_pt['x'],
                                                                df_pt['y'],
                                                                df_pt['z']]))

        df_pt["x"], df_pt["y"] = self.proj_to_proj_out(df_pt['x'], df_pt['y'])

        return df_pt



"""
Acquisition data class module.
"""
import numpy as np
from scipy.spatial.transform import Rotation as R
from borea.geodesy.proj_engine import ProjEngine
from borea.geodesy.approx_euclidean_proj import ApproxEuclideanProj
from borea.geodesy.local_euclidean_proj import LocalEuclideanProj
from borea.utils.check.check_order_axe import check_order_axe


# pylint: disable=too-many-instance-attributes too-many-arguments too-many-positional-arguments
class Shot:
    """
    Shot class definition.

    Args:
        name_shot (str): Name of the shot.
        pos_shot (numpy.array): Array of coordinate position [X, Y, Z].
        ori_shot (numpy.array): Array of orientation of the shot [Omega, Phi, Kappa] in degree.
        name_cam (str): Name of the camera.
        unit_angle (str): Unit of angle 'degrees', 'radian'.
        linear_alteration (bool): True if z shot is correct of linear alteration.
        order_axe (str): Order of rotation matrix axes.
    """
    def __init__(self, name_shot: str, pos_shot: np.ndarray,
                 ori_shot: np.ndarray, name_cam: str,
                 unit_angle: str, linear_alteration: bool,
                 order_axe: str) -> None:
        self.name_shot = name_shot
        self.pos_shot = pos_shot
        self.pos_shot_eucli = None
        self.linear_alteration = linear_alteration
        self.ori_shot = ori_shot
        self.unit_angle = unit_angle
        self.name_cam = name_cam
        self.order_axe = order_axe
        self.z_nadir = None
        self.co_points = {}
        self.gcp2d = {}
        self.gcp3d = {}
        self.mat_rot = self.set_rot_shot()
        self.mat_rot_eucli = None
        self.projeucli = None
        self.approxeucli = False
        self.f_sys = lambda x_shot, y_shot, z_shot: (x_shot, y_shot, z_shot)
        self.f_sys_inv = lambda x_shot, y_shot, z_shot: (x_shot, y_shot, z_shot)

    @classmethod
    def from_param_euclidean(cls, name_shot: str, pos_eucli: np.ndarray,
                             mat_ori_eucli: np.ndarray, name_cam: str,
                             unit_angle: str, linear_alteration: bool,
                             order_axe: str, approx: bool) -> None:
        """
        Construction of a shot object using the Euclidean position.

        Args:
            name_shot (str): Name of the shot.
            pos_eucli (np.array): Euclidean position of the shot.
            mat_ori_eucli (np.array): Euclidean rotation matrix of the shot.
            name_cam (str): Name of the camera.
            unit_angle (str): Unit of angle 'degrees', 'radian'.
            linear_alteration (bool): True if z shot is correct of linear alteration.
            order_axe (str): Order of rotation matrix axes.
            approx (bool): True if you want to use approx euclidean system.

        Returns:
            Shot: The shot.
        """
        shot = cls(name_shot, np.array([0, 0, 0]), np.array([0, 0, 0]),
                   name_cam, unit_angle, linear_alteration, order_axe)
        shot.pos_shot_eucli = pos_eucli
        shot.approxeucli = approx
        if approx:
            shot.projeucli = ApproxEuclideanProj(pos_eucli[0], pos_eucli[1])
        else:
            shot.projeucli = LocalEuclideanProj(pos_eucli[0], pos_eucli[1])
        unitori = shot.unit_angle == "degree"
        shot.pos_shot = shot.projeucli.eucli_to_world(pos_eucli)
        shot.co_points = {}
        shot.gcp3d = {}
        shot.gcp2d = {}
        shot.mat_rot = shot.projeucli.mat_eucli_to_mat(shot.pos_shot[0], shot.pos_shot[1],
                                                       mat_ori_eucli)
        shot.mat_rot_eucli = mat_ori_eucli
        order_xyz = check_order_axe(order_axe)
        shot.ori_shot = -(R.from_euler("x", np.pi) *
                          R.from_matrix(shot.mat_rot)).as_euler(order_xyz, degrees=unitori)

        shot.f_sys = lambda x_shot, y_shot, z_shot: (x_shot, y_shot, z_shot)
        shot.f_sys_inv = lambda x_shot, y_shot, z_shot: (x_shot, y_shot, z_shot)

        return shot

    def set_rot_shot(self) -> np.ndarray:
        """
        Build the rotation matrix with omega phi kappa.

        Returns:
            np.array: The rotation matrix.
        """
        order_xyz = check_order_axe(self.order_axe)
        rot = R.from_euler(order_xyz, -np.array(self.ori_shot), degrees=self.unit_angle == "degree")
        rot = R.from_euler("x", np.pi) * rot
        return rot.as_matrix()

    def set_param_eucli_shot(self, approx: bool) -> None:
        """
        Setting up Euclidean parameters projeucli, pos_shot_eucli, mat_rot_eucli.

        Args:
            approx (bool): True if you want to use approx euclidean system.
        """
        if approx:
            self.projeucli = ApproxEuclideanProj(self.pos_shot[0], self.pos_shot[1])
        else:
            self.projeucli = LocalEuclideanProj(self.pos_shot[0], self.pos_shot[1])

        self.mat_rot_eucli = self.projeucli.mat_to_mat_eucli(self.pos_shot[0],
                                                             self.pos_shot[1],
                                                             self.mat_rot)
        self.approxeucli = approx
        self.pos_shot_eucli = self.projeucli.world_to_eucli(self.pos_shot)

    def set_z_nadir(self, z_nadir: float) -> None:
        """
        Give z nadir to the shot.

        Args:
            z_nadir (flaot): z_nadir of the shot.
        """
        self.z_nadir = z_nadir

    def set_unit_angle(self, unit_angle: str) -> None:
        """
        Allows you to change the orientation angle unit.

        Args:
            unit_angle (str): Unit angle.
        """
        if unit_angle != self.unit_angle:
            self.unit_angle = unit_angle
            if unit_angle == "radian":
                self.ori_shot = self.ori_shot*np.pi/180
            else:
                self.ori_shot = self.ori_shot*180/np.pi

    def set_type_z(self, type_z: str) -> None:
        """
        Allows you to change the type of z.

        Args:
            type_z (str): z type height or altitude.
        """
        if type_z == "height":
            self.pos_shot[2] = ProjEngine().tf.tranform_height(self.pos_shot)
        else:
            self.pos_shot[2] = ProjEngine().tf.tranform_altitude(self.pos_shot)

    def set_linear_alteration(self, linear_alteration: bool) -> None:
        """
        Allows you to correct or de-correct the linear alteration.

        Args:
            linear_alteration (bool): Linear alteration boolean.
        """
        if linear_alteration != self.linear_alteration:
            self.linear_alteration = linear_alteration
            if linear_alteration:
                self.pos_shot[2] = self.get_z_add_scale_factor()
            else:
                self.pos_shot[2] = self.get_z_remove_scale_factor()

    def set_order_axe(self, order_axe: str) -> None:
        """
        Allows to change order of axe for the matrice and angle.

        Args:
            order_axe (str): New order axe.
        """
        if order_axe != self.order_axe:
            self.order_axe = order_axe
            order_xyz = check_order_axe(self.order_axe)
            unitori = self.unit_angle == "degree"
            self.ori_shot = -(R.from_euler("x", np.pi) *
                              R.from_matrix(self.mat_rot)).as_euler(order_xyz, degrees=unitori)

    def set_proj_pos(self) -> None:
        """
        Change the projection of the position shot with the epsg_output.
        """
        self.pos_shot = np.array(ProjEngine().tf.proj_to_proj_out(*self.pos_shot))

    def get_z_remove_scale_factor(self) -> float:
        """
        Return Z after removing the scale factor. The Z of the object is NOT modified.

        Returns:
            float: z without linear alteration.
        """
        if self.z_nadir:
            scale_factor = ProjEngine().get_scale_factor(self.pos_shot[0],
                                                         self.pos_shot[1])
            new_z = (self.pos_shot[2] + scale_factor * self.z_nadir) / (1 + scale_factor)
        else:
            raise ValueError(f"No removing linear alteration of the z shot {self.name_shot}, "
                             "because no dtm or no set z_nadir of shot. For setting z_nadir "
                             "of shots make dtm or add work.set_z_nadir_shot()")

        return new_z

    def get_z_add_scale_factor(self) -> float:
        """
        Return Z after adding the scale factor. The Z of the object is NOT modified.

        Returns:
            float: z with linear alteration.
        """
        if self.z_nadir:
            scale_factor = ProjEngine().get_scale_factor(self.pos_shot[0],
                                                         self.pos_shot[1])
            new_z = self.pos_shot[2] + scale_factor * (self.pos_shot[2] - self.z_nadir)
        else:
            raise ValueError(f"No adding linear alteration of the z shot {self.name_shot}, "
                             "because no dtm or no set z_nadir of shot. For setting z_nadir "
                             "of shots make dtm or add work.set_z_nadir_shot()")

        return new_z



"""
Module for class singleton
"""


class Singleton(type):
    """
    Class Singleton for Dtm and ProjEngine.
    It's a design pattern, belonging to the category of creation patterns,
    whose aim is to restrict the instantiation of a class to a single object.
    """
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]

    def clear(cls):
        """
        Clear instances
        """
        _ = cls._instances.pop(cls, None)



"""
World image transformation module for self
"""
import numpy as np


class WorldImageDtm:
    """
    Function world_to_image and image_to_world for dtm.
    Class parent of dtm.
    """
    def __init__(self, gt: tuple) -> None:
        self.gt = gt

    def image_to_world(self, coor_img: np.ndarray) -> np.ndarray:
        """
        Compute world coordinates from image coordinates.

        Args:
            coor_img (np.array): Coordinate image [column, line].

        Returns:
            np.array: x, y world coordinates.
        """
        if self.gt:
            x = (np.array(coor_img[0])+0.5) * self.gt[1] + self.gt[0]
            y = (np.array(coor_img[1])+0.5) * self.gt[5] + self.gt[3]
        else:
            x, y = np.nan, np.nan
        return np.array([x, y])

    def world_to_image(self, coor_world: np.ndarray) -> np.ndarray:
        """
        Compute image coordinates from world coordinates.

        Args:
            coor_world (np.array): Coordinate world 2D [X, Y].

        Returns:
            np.array: Image coordinates.
        """
        if self.gt:
            col = (np.array(coor_world[0]) - self.gt[0])/self.gt[1] - 0.5
            line = (np.array(coor_world[1]) - self.gt[3])/self.gt[5] - 0.5
        else:
            col, line = np.nan, np.nan
        return np.array([col, line])



"""
Args of parser to write file.
"""
import argparse


def args_writer(parser: argparse) -> argparse:
    """
    Args to write file.

    Args:
        parser (argparse): Parser to add argument.

    Returns:
        argsparse: Parser with argument.
    """
    parser.add_argument('-n', '--namereturn',
                        type=str,
                        help='Name of the new file.')
    parser.add_argument('-w', '--pathreturn',
                        type=str, default='./',
                        help='Conversion path e.g. test/tmp/.')
    return parser



"""
Main to convert opk file to an other opk file
"""
# pylint: disable=import-error, wrong-import-position, line-too-long
import argparse
import sys
import os

sys.path.append(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))
from borea.process.p_format.p_read_opk import args_reading_opk, process_args_read_opk  # noqa: E402
from borea.process.p_format.p_write_opk import args_writing_opk, process_args_write_opk  # noqa: E402, E501


def opk_to_opk():
    """
    Converts an OPK file into an OPK file.
    """
    parser = argparse.ArgumentParser(description='Photogrammetric site conversion'
                                                 ' and manipulation software opk to opk.')
    # Args for implement opk to opk
    parser = args_reading_opk(parser)
    parser = args_writing_opk(parser)

    args = parser.parse_args()

    # Process to read data
    work = process_args_read_opk(args)
    # Process to write opk
    process_args_write_opk(args, work)


if __name__ == "__main__":
    opk_to_opk()



"""
Args of parser to calcule 6 externals parameters of shots.
"""
import argparse

import numpy as np
from borea.process.p_add_data.p_file_gcp2d import args_gcp2d
from borea.process.p_add_data.p_gen_param import args_general_param, process_args_gen_param
from borea.process.p_add_data.p_pt3d import args_add_pt3d
from borea.reader.reader_point import read_file_pt_dataframe
from borea.transform_world_image.transform_worksite.space_resection import SpaceResection
from borea.utils.check.check_args_opk import check_header_file
from borea.worksite.worksite import Worksite
from borea.process.p_add_data.p_file_gcp3d import args_gcp3d


def args_space_resection(parser: argparse) -> argparse:
    """
    Args to calcule 6 externals parameters of shots.

    Args:
        parser (argparse): Parser to add argument.

    Returns:
        argsparse: Parser with argument.
    """
    parser = args_add_pt3d(parser)
    parser = args_general_param(parser)
    parser = args_gcp2d(parser)
    parser = args_gcp3d(parser)
    return parser


def process_space_resection(args: argparse) -> Worksite:
    """
    Processing args with data.

    Args:
        args (argparse): Arg to apply on worksite (data).

    Returns:
        Worksite: data
    """
    work = Worksite("Space_Resection")
    work = process_args_gen_param(args, work)
    work.set_type_z_shot(check_header_file(list(args.output_header.upper()))[1])
    pt2d, _ = read_file_pt_dataframe(args.gcp2d, list(args.head_gcp2d.upper()), "pt2d")
    pt3d, _ = read_file_pt_dataframe(args.gcp3d, list(args.head_gcp3d.upper()), "pt3d")
    pinit = {"coor_init": np.array(args.point3d)}
    SpaceResection(work).space_resection_to_worksite(pt2d, pt3d, pinit)
    return work



"""
Image world transformation module for Shot
"""
import numpy as np
from borea.datastruct.shot import Shot
from borea.datastruct.camera import Camera
from borea.datastruct.dtm import Dtm
from borea.geodesy.proj_engine import ProjEngine
from borea.transform_world_image.transform_shot.conversion_coor_shot import conv_z_shot_to_z_data
from borea.transform_world_image.transform_shot.conversion_coor_shot import conv_output_z_type


class ImageWorldShot():
    """
    Function image_to_world for shot.

    Args:
        shot (Shot): The shot for convert coordinate.
        cam (Camera): The camera of the shot.
    """
    def __init__(self, shot: Shot, cam: Camera) -> None:
        self.shot = shot
        self.cam = cam

    def image_to_world(self, img_coor: np.ndarray,
                       type_z_data: str, type_z_shot: str,
                       nonadir: bool = True) -> np.ndarray:
        """
        Calculate x and y cartographique coordinate with z.

        Args:
            img_coor (np.array): Image coordinate [col line].
            type_z_data (str): Type of z data you want in output, "height" or "altitude".
            type_z_shot (str): Type of z shot, "height" or "altitude".
            nonadir (bool): To calculate nadir no take linear alteration.

        Returns:
            np.array: Cartographique coordinate [x,y,z].
        """
        if type_z_data != type_z_shot and not ProjEngine().tf.geog_to_geoid:
            raise ValueError("Missing geoid")

        if not Dtm().path_dtm:
            raise ValueError("Missing dtm")

        img_coor = np.squeeze(img_coor)
        coor_world = self.image_world_iter(img_coor, type_z_shot, nonadir)

        coor_world = conv_output_z_type(coor_world, Dtm().type_dtm, type_z_data)

        return coor_world

    def image_world_iter(self, img_coor: np.ndarray,
                         type_z_shot: str, nonadir: bool = True) -> np.ndarray:
        """
        Calculate x and y cartographique coordinate with z.

        Args:
            img_coor (np.array): Image coordinate [col line].
            type_z_shot (str): Type of z shot, "height" or "altitude".
            nonadir (bool): To calculate nadir no take linear alteration.

        Returns:
            np.array: Cartographique coordinate [x,y,z].
        """
        z_world = np.full_like(img_coor[0], Dtm().get_z_world(self.shot.pos_shot[0:2]))
        coor_world = self.image_z_to_world(img_coor, type_z_shot, z_world, nonadir)
        precision_reached = False
        nbr_iter = 0
        iter_max = 10
        while not precision_reached and nbr_iter < iter_max:
            z_world = np.squeeze(Dtm().get_z_world(coor_world[0:2]))
            coor_new_world = self.image_z_to_world(img_coor, type_z_shot, z_world, nonadir)
            x_diff = (coor_new_world[0] - coor_world[0]) ** 2
            y_diff = (coor_new_world[1] - coor_world[1]) ** 2
            z_diff = (coor_new_world[2] - coor_world[2]) ** 2
            dist2 = x_diff + y_diff + z_diff
            precision_reached = np.any(dist2 < 0.01**2)
            coor_world = coor_new_world
            nbr_iter += 1

        return coor_world

    def image_z_to_world(self, img_coor: np.ndarray, type_z_shot: str,
                         z: np.ndarray = 0, nonadir: bool = True) -> np.ndarray:
        """
        Calculate x and y cartographique coordinate with z.

        Args:
            img_coor (np.array): Image coordinate [col line].
            type_z_shot (str): Type of z "height" or "altitude".
            z (Union[np.array, float]): La position z du point par défault = 0.
            nonadir (bool): To calculate nadir no take linear alteration.

        Returns:
            np.array: Cartographique coordinate [x,y,z].
        """
        if isinstance(img_coor[0], np.ndarray):
            if np.all(z == 0):
                z = np.full_like(img_coor[0], 0)

        pt_bundle = self.image_to_bundle(img_coor)

        pos_shot_new_z = conv_z_shot_to_z_data(self.shot, type_z_shot, Dtm().type_dtm,
                                               nonadir, self.shot.approxeucli)

        pos_eucli = self.shot.projeucli.world_to_eucli(pos_shot_new_z)

        pt_eucli = self.local_to_eucli(pt_bundle, pos_eucli, z)

        pt_world = self.shot.projeucli.eucli_to_world(pt_eucli)
        np.array([pt_world[0], pt_world[1], z])

        return np.array([pt_world[0], pt_world[1], z])

    def image_to_bundle(self, img_coor: np.ndarray) -> np.ndarray:
        """
        Convert coordinate image col line to coordinate bundle.

        Args:
            img_coor (np.array): Image coordinate [col line].

        Returns:
            np.array: Cartographique coordinate [x,y,z].
        """
        x_shot = img_coor[0] - self.cam.ppax
        y_shot = img_coor[1] - self.cam.ppay
        z_shot = np.full_like(x_shot, self.cam.focal)
        x_shot, y_shot, z_shot = self.shot.f_sys_inv(x_shot, y_shot, z_shot)
        x_bundle = x_shot / self.cam.focal * z_shot
        y_bundle = y_shot / self.cam.focal * z_shot
        z_bundle = z_shot
        return np.array([x_bundle, y_bundle, z_bundle])

    def local_to_eucli(self, pt_bundle: np.ndarray, pos_eucli: np.ndarray,
                       z: np.ndarray) -> np.ndarray:
        """
        Convert coordinate point in local system to euclidean system.

        Args:
            pt_bundle (np.array): Point to convert [X, Y, Z].
            pos_eucli (np.array): Shot position in the euclidean system.
            z (np.array): Z position of points.

        Returns:
            np.array: [X, Y, Z] in euclidean system.
        """
        p_local = self.shot.mat_rot_eucli.T @ np.vstack(pt_bundle)
        p_local = np.squeeze(p_local + pos_eucli.reshape((3, 1)))
        lamb = (z - pos_eucli[2])/(p_local[2] - pos_eucli[2])
        x_local = pos_eucli[0] + (p_local[0] - pos_eucli[0]) * lamb
        y_local = pos_eucli[1] + (p_local[1] - pos_eucli[1]) * lamb
        return np.array([x_local, y_local, z])



"""
Args of parser to calcule image coordinate.
"""
import argparse
from borea.process.p_add_data.p_file_gcp2d import args_gcp2d, process_gcp2d
from borea.process.p_add_data.p_write import args_writer
from borea.worksite.worksite import Worksite
from borea.process.p_add_data.p_file_gcp3d import args_gcp3d, process_gcp3d
from borea.transform_world_image.transform_worksite.world_image_work import WorldImageWork
from borea.writer.writer_df_to_txt import write_df_to_txt


def args_world_image(parser: argparse) -> argparse:
    """
    Args to control photogrametrique file.

    Args:
        parser (argparse): Parser to add argument.

    Returns:
        argsparse: Parser with argument.
    """
    parser = args_gcp3d(parser)
    parser = args_gcp2d(parser)
    parser.add_argument('-d', '--control_type',
                        type=str, default=[], nargs='*',
                        help='Type of gcp to control.')
    parser = args_writer(parser)
    return parser


def process_world_image(args, work: Worksite) -> None:
    """
    Processing args with data.

    Args:
        args (argparse): Arg to apply on worksite (data).
        work (Worksite): Data.
    """
    # Read GCP 3D and 2D
    work = process_gcp3d(args, work)
    work = process_gcp2d(args, work)

    # Calculate image coordinate of GCP if they exist
    print("Calculation of data world to image.")
    WorldImageWork(work).calculate_world_to_image(args.control_type)

    df2d = work.get_point_image_dataframe("gcp3d", [])
    write_df_to_txt(args.namereturn, args.pathreturn, df2d)



"""
Args of parser to calcule world coordinate.
"""
import argparse
from borea.process.p_add_data.p_write import args_writer
from borea.worksite.worksite import Worksite
from borea.process.p_add_data.p_file_gcp2d import args_gcp2d, process_gcp2d
from borea.transform_world_image.transform_worksite.image_world_work import ImageWorldWork
from borea.writer.writer_df_to_txt import write_df_to_txt


def args_image_world(parser: argparse) -> argparse:
    """
    Args to control photogrametrique file.

    Args:
        parser (argparse): Parser to add argument.

    Returns:
        argsparse: Parser with argument.
    """
    parser = args_gcp2d(parser)
    parser.add_argument('-p', '--process',
                        type=str, default="inter", choices=["inter", "square"],
                        help="Type of process for the function image to world,"
                             "intersection or least_square")
    parser = args_writer(parser)
    return parser


def process_image_world(args, work: Worksite) -> None:
    """
    Processing args with data.

    Args:
        args (argparse): Arg to apply on worksite (data).
        work (Worksite): Data.
    """
    # Read GCP 2D
    work = process_gcp2d(args, work)

    # Calculate ground coordinate of conneting point by intersection
    print("Calculation of data image to world.")
    ImageWorldWork(work).manage_image_world(type_point="gcp2d",
                                            type_process=args.process,
                                            control_type=None)
    df3d = work.get_point_world_dataframe("gcp2d", [])
    write_df_to_txt(args.namereturn, args.pathreturn, df3d)



"""
Worksite data class module.
"""
import numpy as np
import pandas as pd
from borea.datastruct.dtm import Dtm
from borea.datastruct.workdata import Workdata
from borea.geodesy.proj_engine import ProjEngine
from borea.transform_world_image.transform_shot.image_world_shot import ImageWorldShot


class Worksite(Workdata):
    """
    Worksite class, class main of the tools.

    Args:
        name (str): Name of the worksite.
    """

    def get_point_image_dataframe(self, type_point: str, control_type: list = None) -> pd.DataFrame:
        """
        Retrieves id_pt, id_img, column, line in a pandas of the requested point type.

        Args:
            type_point (str): "co_points" or "gcp2d" depending on what you want to get.
            control_type (list): Type controle for gcp.

        Returns:
            pandas: Pandas table.
        """
        if type_point not in ["co_points", "gcp2d", "gcp3d"]:
            raise ValueError(f"type_point {type_point} is incorrect,['co_points','gcp2d','gcp3d']")

        if type_point == "co_points" or not control_type:
            control_type = []

        type_iter = type_point if type_point != "gcp3d" else "gcp2d"

        id_pt = []
        id_img = []
        coor = []
        for name_pt, list_shot in getattr(self, type_iter).items():
            if control_type and self.gcp3d[name_pt].code not in control_type:
                continue
            for name_shot in list_shot:
                id_pt.append(name_pt)
                id_img.append(name_shot)
                coor.append(getattr(self.shots[name_shot], type_point)[name_pt])

        coor = np.array(coor)
        return pd.DataFrame({"id_pt": id_pt, "id_img": id_img,
                             "column": coor[:, 0], "line": coor[:, 1]})

    def set_point_image_dataframe(self, pd_mes: pd.DataFrame, type_point: str) -> None:
        """
        Set requested point type by pandas table id_pt, id_img, column, line.

        Args:
            pd_mes (pd): Pandas table of data.
            type_point (str): "co_points" or "gcp2d" depending on what you want to set.
        """
        if type_point not in ["co_points", "gcp2d"]:
            raise ValueError(f"type_point {type_point} is incorrect,['co_points','gcp2d']")

        for _, row in pd_mes.iterrows():
            try:
                getattr(self.shots[row['id_img']],
                        type_point)[row['id_pt']] = np.array([row['column'], row['line']])
            except KeyError:
                continue

    def get_point_world_dataframe(self, type_point: str, control_type: list) -> pd.DataFrame:
        """
        Retrieves id_pt, x, y, z in a pandas of the requested point type.

        Args:
            type_point (str): "co_points" or "gcp2d" depending on what you want to get.
            control_type (list): Type controle for gcp.

        Returns:
            pandas: Pandas table.
        """
        if type_point not in ["co_points", "gcp2d"]:
            raise ValueError(f"type_point {type_point} is incorrect,['co_points','gcp2d']")

        out_pt, control_type = self.get_attr_transfo_pt(type_point, control_type)

        id_pt = []
        coor = []
        for name_pt, coor_pt in getattr(self, out_pt).items():
            if control_type and self.gcp3d[name_pt].code not in control_type:
                continue
            id_pt.append(name_pt)
            coor.append(coor_pt)

        coor = np.array(coor)
        return pd.DataFrame({"id_pt": id_pt, "x": coor[:, 0], "y": coor[:, 1], "z": coor[:, 2]})

    def set_point_world_dataframe(self, pd_mes: pd.DataFrame, type_point: str) -> None:
        """
        Set requested point type by pandas table id_pt, x, y, z.

        Args:
            pd_mes (pd): Pandas table of data.
            type_point (str): "co_points" or "gcp2d" depending on what you want to set.
        """
        if type_point not in ["co_points", "gcp2d"]:
            raise ValueError(f"type_point {type_point} is incorrect,['co_points','gcp2d']")

        out_pt, _ = self.get_attr_transfo_pt(type_point, None)

        if "type" not in list(pd_mes.columns):
            for _, row in pd_mes.iterrows():
                getattr(self, out_pt)[row['id_pt']] = np.array([row['x'], row['y'], row['z']])
        else:
            for _, row in pd_mes.iterrows():
                self.add_gcp3d(row['id_pt'], row['type'], np.array([row['x'], row['y'], row['z']]))

    def get_coor_pt_img_and_world(self, name_shot: str, type_point: str) -> tuple:
        """
        Recovers image and terrain data for all link points in the image.

        Args:
            name_shot (str): Name of shot to take.
            type_point (str): Type point to take "co_points" or "gcp2d".

        Returns:
            tuple: np.array(obs_image), np.array(pt_world).
        """
        if type_point not in ['co_points', 'gcp2d']:
            raise ValueError(f"type point {type_point} is incorrect.['co_points','gcp2d']")

        out_pt, _ = self.get_attr_transfo_pt(type_point, None)

        if not getattr(self, out_pt):
            raise ValueError(f"Attribut {out_pt} in worksite is empty.")

        co_point = getattr(self.shots[name_shot], type_point)
        world_pt = []
        img_pt = []
        for key_pt, img_coor in co_point.items():
            try:
                pt_world = getattr(self, out_pt)[key_pt]
                world_pt.append(pt_world)
                img_pt.append(img_coor)
            except KeyError:
                print(f"Point {key_pt}, doesn't have world coordinate, perhaps a single point.")

        return np.stack(img_pt, axis=-1), np.stack(world_pt, axis=-1)

    def set_param_shot(self, approx=False) -> None:
        """
        Setting up different parameters in shot by data ressources.

        Args:
            approx (bool): True if you want to use approx euclidean system.
        """
        check_dtm = True
        if not Dtm().path_dtm:
            check_dtm = False

        if not ProjEngine().geoid:
            raise ValueError("you have not entered all the information required to set up"
                             " the projection system. Because type z shot != type z data.")

        self.approxeucli = approx
        for shot in self.shots.values():
            shot.set_param_eucli_shot(approx)
            if check_dtm:
                cam = self.cameras[shot.name_cam]
                z_nadir = ImageWorldShot(shot, cam).image_to_world(np.array([cam.ppax, cam.ppay]),
                                                                   self.type_z_shot,
                                                                   self.type_z_shot, False)[2]
                shot.set_z_nadir(z_nadir)

    # pylint: disable-next=too-many-arguments too-many-positional-arguments
    def set_unit_output(self, type_z: str = None, unit_angle: str = None,
                        linear_alteration: bool = None, order_axe: str = None,
                        proj_output: bool = True) -> None:
        """
        Allows you to change unit or parameter of shots.

        Args:
            type_z (str): Unit of z shot you want.
            unit_angle (str): Unit angle you want.
            linear_alteration (bool): True if you want data corrected.
            order_axe (str): Order of rotation matrice you want in your angle.
            proj_output (bool): True to change the projection of shot.
        """
        if unit_angle not in ["degree", "radian", None]:
            raise ValueError(f"unit_angle: {unit_angle} is not recognized,"
                             "recognized values are degree and radian.")

        if type_z not in ["height", "altitude", None]:
            raise ValueError(f"type_z: {type_z} is not recognized,"
                             "recognized values are altitude and height.")

        if type_z == self.type_z_shot:
            type_z = None
        else:
            self.type_z_shot = type_z

        for shot in self.shots.values():
            if unit_angle:
                shot.set_unit_angle(unit_angle)
            if type_z:
                shot.set_type_z(type_z)
            if linear_alteration is not None:
                shot.set_linear_alteration(linear_alteration)
            if order_axe:
                shot.set_order_axe(order_axe)
            if self.epsg_output and proj_output:
                shot.set_proj_pos()

    def calculate_barycentre(self) -> np.ndarray:
        """
        Calculate barycentre of the worksite.

        Returns:
            np.array: The barycentre [X, Y, Z].
        """
        size = len(self.shots)
        pos = np.zeros((size, 3))
        i = 0
        for shot in self.shots.values():
            pos[i, :] = shot.pos_shot
            i += 1
        return np.mean(pos, axis=0)

    def get_attr_transfo_pt(self, type_point: str, control_type: list) -> tuple:
        """
        Get string name of attribut where is save transformation image to world point.

        Args:
            type_point (str):"co_points" or "gcp2d"
                              depending on what you want.
            control_type (list): Type controle for gcp.
        """
        if type_point == 'co_points':
            out_pt = "co_pts_world"
            control_type = []
        else:
            out_pt = "gcp2d_in_world"

        return out_pt, control_type



"""
Args of parser for writing Rpc file
"""
import argparse
from borea.worksite.worksite import Worksite
from borea.writer.manage_writer import manager_writer


def args_writing_rpc(parser: argparse) -> argparse:
    """
    Args for writing rpc file.

    Args:
        parser (argparse): Parser to add argument.

    Returns:
        argsparse: Parser with argument.
    """
    parser.add_argument('-w', '--pathreturn',
                        type=str, default='./',
                        help='Conversion path e.g. test/tmp/.')
    parser.add_argument('-o', '--order',
                        type=int, default=3, choices=[1, 2, 3],
                        help="Degree of the polynomial of the rpc (1, 2, 3)")
    parser.add_argument('-d', '--size_grid',
                        type=int, default=100,
                        help="Size of the grid to calculate Rpc.")
    parser.add_argument('-l', '--fact_rpc',
                        type=float, default=None,
                        help="Factor Rpc for pyproj convertion.")
    return parser


def process_args_write_rpc(args: argparse, work: Worksite) -> None:
    """
    Processing args with data.

    Args:
        args (argparse): Arg to apply on worksite (data).
        work (Worksite): Data.
    """
    # Writing data
    print("Writing Rpc.")
    args_writing = {"order": args.order,
                    "size_grid": args.size_grid,
                    "fact_rpc": args.fact_rpc}
    manager_writer("rpc", None, args.pathreturn, args_writing, work)
    print(f"File written in folder {args.pathreturn} .txt.")



"""
Main to calculate world coodinate of the image point.
"""
# pylint: disable=import-error, wrong-import-position
import argparse
import sys
import os

sys.path.append(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))
from borea.process.p_add_data.p_add_shot import args_add_shot, process_add_shot  # noqa: E402
from borea.process.p_add_data.p_pt3d import args_add_pt3d, process_world_image  # noqa: E402


def pt_world_to_image():
    """
    Converts a terrain point into an image point.
    """
    parser = argparse.ArgumentParser(description='Calculate world coodinate of the image point.')
    # Args for implement pt world to image
    parser = args_add_shot(parser)
    parser = args_add_pt3d(parser)

    args = parser.parse_args()

    # Process to read data
    work = process_add_shot(args)
    # Process to transform world to image
    process_world_image(args, work)


if __name__ == "__main__":
    pt_world_to_image()



"""
Args of parser for reading generals parameters
"""
import argparse
from borea.worksite.worksite import Worksite
from borea.process.p_add_data.p_proj import args_proj_param, process_args_proj_param
from borea.reader.reader_camera import read_camera


def args_general_param(parser: argparse) -> argparse:
    """
    Args for adding general parameter.

    Args:
        parser (argparse): Parser to add argument.

    Returns:
        argsparse: Parser with argument.
    """
    parser = args_proj_param(parser)
    parser.add_argument('-c', '--camera',
                        type=str, nargs='*',
                        help='Files paths of cameras (xml or txt).')
    parser.add_argument('-m', '--dtm',
                        type=str, default=None,
                        help='DtM of the worksite.')
    parser.add_argument('--fm', '--format_dtm',
                        type=str, default=None,
                        help='Format of Dtm "altitude" or "height".')
    parser.add_argument('-x', '--approx_system',
                        type=bool, default=False,
                        help="To use an approximate system.")
    return parser


def process_args_gen_param(args: argparse, work: Worksite) -> Worksite:
    """
    Processing args with data.

    Args:
        args (argparse): Arg to apply on worksite (data).
        work (Worksite): Worksite to work on.

    Returns:
        Worksite: data
    """
    # Add a projection to the worksite
    work = process_args_proj_param(args, work)

    # Reading camera file
    if args.camera is not None:
        read_camera(args.camera, work)
        print(f"Camera file reading done. {len(args.camera)} read")
    else:
        print("There is no given camera.")

    # Add Dem
    work.set_dtm(args.dtm, args.fm)
    if args.dtm is not None:
        print("Add dtm to the worksite done.")
    else:
        print("Not Dtm in the worksite.")

    work.set_param_shot(args.approx_system)
    return work



"""
Script to read point (connecting point, gcp2d gcp3d) format
.txt/.mes/.app with data arranged in columns.
"""
from pathlib import Path, PureWindowsPath
import pandas as pd
import numpy as np
from borea.worksite.worksite import Worksite
from borea.utils.check.check_args_reader_pt import check_header_file


def read_file_pt(path: str, header: list, type_point: str, work: Worksite) -> None:
    """
    Read file of points.

    Agrs:
        file (str): Path of points file.
        header (list): Header of file to read column.
        type_point (str): Type of point is reading (co_point, gcp2d, gcp3d).
        work (Worksite): Worksite which needs connecting points.
    """
    if type_point not in ["co_point", "gcp2d", "gcp3d"]:
        raise ValueError(f"type {type_point} in incorrect. ['co_point', 'gcp2d', 'gcp3d']")

    header, type_z = check_header_file(header, type_point)
    work.type_z_data = type_z

    try:
        with open(Path(PureWindowsPath(path)), 'r', encoding="utf-8") as file_pts:
            for pt in file_pts.readlines():
                if pt != '\n' and pt[0] != '#':
                    info = pt.split()

                    if type_point == "gcp3d":
                        coor = np.array([float(info[header.index("X")]),
                                         float(info[header.index("Y")]),
                                         float(info[header.index("Z")])])
                        type_2args = "T"
                    else:
                        coor = np.array([float(info[header.index("X")]),
                                         float(info[header.index("Y")])])
                        type_2args = "N"

                    try:
                        getattr(work, "add_" + type_point)(info[header.index("P")],
                                                           info[header.index(type_2args)],
                                                           coor)
                    except ValueError as e:
                        raise ValueError("The letter T is missing "
                                         "from the header of the file.") from e
            file_pts.close()
    except FileNotFoundError as e:
        raise FileNotFoundError(f"The path {path} is incorrect !!!") from e


def read_file_pt_dataframe(path: str, header: list, type_point: str) -> tuple:
    """
    Read file of points to save in Dataframe.

    Agrs:
        file (str): Path of points file.
        header (list): Header of file to read column.
        type_point (str): Type of point is reading (pt2d, pt3d).

    Returns:
        tuple: Dataframe of data and type of z in str.
    """
    if type_point not in ["pt2d", "pt3d"]:
        raise ValueError(f"type {type_point} in incorrect. ['pt2d', 'pt3d']")

    header, type_z = check_header_file(header, type_point)

    id_pt = []
    ttype = []
    coor = []
    try:
        with open(Path(PureWindowsPath(path)), 'r', encoding="utf-8") as file_pts:
            for pt in file_pts.readlines():
                if pt != '\n' and pt[0] != '#':
                    info = pt.split()

                    if type_point == "pt3d":
                        coor.append([float(info[header.index("X")]),
                                     float(info[header.index("Y")]),
                                     float(info[header.index("Z")])])
                        type_2args = "T"
                    else:
                        coor.append([float(info[header.index("X")]),
                                     float(info[header.index("Y")])])
                        type_2args = "N"

                    id_pt.append(info[header.index("P")])
                    try:
                        ttype.append(info[header.index(type_2args)])
                    except ValueError:
                        continue
            file_pts.close()
    except FileNotFoundError as e:
        raise FileNotFoundError(f"The path {path} is incorrect !!!") from e

    coor = np.array(coor)
    if type_point == "pt3d":
        df = pd.DataFrame({"id_pt": id_pt,
                           "type": ttype if ttype else None,
                           "x": coor[:, 0],
                           "y": coor[:, 1],
                           "z": coor[:, 2]})
    else:
        df = pd.DataFrame({"id_pt": id_pt,
                           "id_shot": ttype,
                           "column": coor[:, 0],
                           "line": coor[:, 1]})
    return df, type_z



"""
Main to control opk file
"""
# pylint: disable=import-error, wrong-import-position
import argparse
import sys
import os

sys.path.append(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))
from borea.process.p_format.p_read_opk import args_reading_opk, process_args_read_opk  # noqa: E402
from borea.process.p_func.p_control import args_control, process_args_control  # noqa: E402


def opk_control():
    """
    Controls the accuracy of an OPK file with support points.
    """
    parser = argparse.ArgumentParser(description='Photogrammetric site control opk file')

    # Args for implement control opk
    parser = args_reading_opk(parser)
    parser = args_control(parser)

    args = parser.parse_args()

    # Process to read data
    work = process_args_read_opk(args)
    # Process to control data
    process_args_control(args, work)


if __name__ == "__main__":
    opk_control()



"""
Function to implement xml.
"""
import xml.etree.ElementTree as ET
import numpy as np


def format_xml(elem: any, prec: str = ".5f") -> str:
    """
    Write item in xml format.

    Args:
        elem (any): The element to write in xml format.
        prec (str): Precision of the element if is a number.

    Returns:
        str: elem in xml format.
    """
    if isinstance(elem, float) and np.isnan(elem):
        return "0.00000"
    if isinstance(elem, (float, np.float32, np.float64)):
        return f"{elem:{prec}}"
    if elem is None:
        return ""

    return f"{elem}"


def indent(elem: ET, level: int = 0) -> None:
    """
    Structure the xml file.

    Args:
        elem (ET): The xml to write.
        level (int): Level of structure.
    """
    i = "\n" + level * "  "
    if len(elem):
        if not elem.text or not elem.text.strip():
            elem.text = i + "  "
        if not elem.tail or not elem.tail.strip():
            elem.tail = i
        for subelem in elem:
            indent(subelem, level + 1)
        if not elem.tail or not elem.tail.strip():
            elem.tail = i
    else:
        if level and (not elem.tail or not elem.tail.strip()):
            elem.tail = i


def add_elem(xml: ET, balises: list, vals: list, prec: str = ".5f") -> None:
    """
    Add a list of element to the ET.

    Args:
        xml (ET): The xml to add elem.
        balises (list): List of name of the balise.
        vals (list): List of item to put on.
        prec (str): Precision of the element if is a number.
    """
    for balise, val in zip(balises, vals):
        ET.SubElement(xml, balise).text = format_xml(val, prec)



"""
Args of parser to transform projection of 3D points.
"""
import argparse

from borea.process.p_add_data.p_file_gcp3d import args_gcp3d
from borea.process.p_add_data.p_proj import args_proj_param
from borea.process.p_add_data.p_write import args_writer
from borea.reader.reader_point import read_file_pt_dataframe
from borea.geodesy.proj_engine import ProjEngine
from borea.writer.writer_df_to_txt import write_df_to_txt


def args_tf_proj_param(parser: argparse) -> argparse:
    """
    Args for adding transform proj of points parameter.

    Args:
        parser (argparse): Parser to add argument.

    Returns:
        argsparse: Parser with argument.
    """
    parser = args_gcp3d(parser)
    parser = args_proj_param(parser)
    parser.add_argument('--oz', '--z_output',
                        type=str, choices=[None, 'altitude', 'height'], default=None,
                        help="Output type of z. altitude or height")
    parser = args_writer(parser)
    return parser


def process_tf_proj_param(args: argparse) -> None:
    """
    Processing args with data.

    Args:
        args (argparse): Arg to apply on worksite (data).
    """
    # Read file to Dataframe
    df, type_z = read_file_pt_dataframe(args.gcp3d, args.head_gcp3d, "pt3d")

    # Setup Projection
    ProjEngine().set_epsg([args.epsg, args.geog, args.geoc], args.pathgeoid, args.oe)

    # Change projection
    new_df = ProjEngine().tf.transform_pt_proj(df, type_z, args.oz)

    # Write the new file
    write_df_to_txt(args.namereturn, args.pathreturn, new_df)



"""
Module math
"""
from scipy.sparse import coo_matrix
import numpy as np


def invert_diag_sparse_matrix_3_3(mat_inv: np.ndarray) -> np.ndarray:
    """
    Block matrix inversion.

    Args:
        mat (np.array): Matrix to flip.

    Returns:
        np.array: Flip matrix.
    """
    mat = np.c_[np.pad(mat_inv.diagonal(-2), (2, 0)),
                np.pad(mat_inv.diagonal(-1), (1, 0)),
                mat_inv.diagonal(0),
                np.pad(mat_inv.diagonal(1), (0, 1)),
                np.pad(mat_inv.diagonal(2), (0, 2))]

    mat[::3, :] = np.roll(mat[::3, :], -2)
    mat[1::3, :] = np.roll(mat[1::3, :], -1)

    data = np.linalg.inv(mat[:, :-2].reshape(1, -1, 3, 3)).flatten()
    coord_i = np.repeat(np.arange(mat_inv.shape[0]), 3)
    coord_j = np.tile(np.arange(mat_inv.shape[0]).reshape(-1, 3), 3).flatten()

    return coo_matrix((data, (coord_i, coord_j)), shape=mat_inv.shape).tocsr()
